# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2016-11-29 07:17
from __future__ import unicode_literals

import autoslug.fields
from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import versionfield


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('flow', '0013_add_requirements'),
    ]

    operations = [
        migrations.CreateModel(
            name='Entity',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', autoslug.fields.AutoSlugField(editable=True, max_length=100, populate_from='name', unique_with=('version',))),
                ('version', versionfield.VersionField(default=0)),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('description', models.TextField(blank=True)),
                ('settings', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('descriptor', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('collections', models.ManyToManyField(to='flow.Collection')),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
                ('data', models.ManyToManyField(to='flow.Data')),
                ('descriptor_schema', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='flow.DescriptorSchema')),
                ('public_processes', models.ManyToManyField(to='flow.Process')),
            ],
            options={
                'default_permissions': (),
                'abstract': False,
                'permissions': (('view_entity', 'Can view entity'), ('edit_entity', 'Can edit entity'), ('share_entity', 'Can share entity'), ('download_entity', 'Can download files from entity'), ('add_entity', 'Can add data objects to entity'), ('owner_entity', 'Is owner of the entity')),
            },
        ),
        migrations.AlterUniqueTogether(
            name='entity',
            unique_together=set([('slug', 'version')]),
        ),
    ]



""".. Ignore pydocstyle D400.

==========
Data Purge
==========

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os
import shutil

from django.conf import settings

from resolwe.flow.models import Data
from resolwe.flow.utils import iterate_fields
from resolwe.utils import BraceMessage as __

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def get_purge_files(root, output, output_schema, descriptor, descriptor_schema):
    """Get files to purge."""
    def remove_file(fn, paths):
        """From paths remove fn and dirs before fn in dir tree."""
        while fn:
            for i in range(len(paths) - 1, -1, -1):
                if fn == paths[i]:
                    paths.pop(i)
            fn, _ = os.path.split(fn)

    def remove_tree(fn, paths):
        """From paths remove fn and dirs before or after fn in dir tree."""
        for i in range(len(paths) - 1, -1, -1):
            head = paths[i]
            while head:
                if fn == head:
                    paths.pop(i)
                    break
                head, _ = os.path.split(head)

        remove_file(fn, paths)

    def subfiles(root):
        """Extend unreferenced list with all subdirs and files in top dir."""
        subs = []
        for path, dirs, files in os.walk(root, topdown=False):
            path = path[len(root) + 1:]
            subs.extend(os.path.join(path, f) for f in files)
            subs.extend(os.path.join(path, d) for d in dirs)
        return subs

    unreferenced_files = subfiles(root)

    remove_file('jsonout.txt', unreferenced_files)
    remove_file('stderr.txt', unreferenced_files)
    remove_file('stdout.txt', unreferenced_files)

    meta_fields = [
        [output, output_schema],
        [descriptor, descriptor_schema]
    ]

    for meta_field, meta_field_schema in meta_fields:
        for field_schema, fields in iterate_fields(meta_field, meta_field_schema):
            if 'type' in field_schema:
                field_type = field_schema['type']
                field_name = field_schema['name']

                # Remove basic:file: entries
                if field_type.startswith('basic:file:'):
                    remove_file(fields[field_name]['file'], unreferenced_files)

                # Remove list:basic:file: entries
                elif field_type.startswith('list:basic:file:'):
                    for field in fields[field_name]:
                        remove_file(field['file'], unreferenced_files)

                # Remove basic:dir: entries
                elif field_type.startswith('basic:dir:'):
                    remove_tree(fields[field_name]['dir'], unreferenced_files)

                # Remove list:basic:dir: entries
                elif field_type.startswith('list:basic:dir:'):
                    for field in fields[field_name]:
                        remove_tree(field['dir'], unreferenced_files)

                # Remove refs entries
                if field_type.startswith('basic:file:') or field_type.startswith('basic:dir:'):
                    for ref in fields[field_name].get('refs', []):
                        remove_tree(ref, unreferenced_files)

                elif field_type.startswith('list:basic:file:') or field_type.startswith('list:basic:dir:'):
                    for field in fields[field_name]:
                        for ref in field.get('refs', []):
                            remove_tree(ref, unreferenced_files)

    return set([os.path.join(root, filename) for filename in unreferenced_files])


def data_purge(data_ids=None, delete=False, verbosity=0):
    """Print files not referenced from meta data.

    If data_ids not given, run on all data objects.
    If delete is True, delete unreferenced files.

    """
    data_path = settings.FLOW_EXECUTOR['DATA_DIR']
    runtime_path = settings.FLOW_EXECUTOR['RUNTIME_DIR']
    unreferenced_files = set()

    data_qs = Data.objects.filter(status__in=[Data.STATUS_DONE, Data.STATUS_ERROR])
    if data_ids is not None:
        data_qs = data_qs.filter(pk__in=data_ids)

    for data in data_qs:
        root = os.path.join(data_path, str(data.id))

        unreferenced_files.update(get_purge_files(
            root,
            data.output,
            data.process.output_schema,
            data.descriptor,
            getattr(data.descriptor_schema, 'schema', [])
        ))

    # Remove any folders, which do not belong to any data objects.
    if data_ids is None:
        for base_path in (data_path, runtime_path):
            for directory in os.listdir(base_path):
                directory_path = os.path.join(base_path, directory)
                if not os.path.isdir(directory_path):
                    continue

                try:
                    data_id = int(directory)
                except ValueError:
                    continue

                # Check if a data object with the given identifier exists.
                if not Data.objects.filter(pk=data_id).exists():
                    unreferenced_files.add(directory_path)

    if verbosity >= 1:
        # Print unreferenced files
        if unreferenced_files:
            logger.info(__("Unreferenced files ({}):", len(unreferenced_files)))
            for name in unreferenced_files:
                logger.info(__("  {}", name))
        else:
            logger.info("No unreferenced files")

    # Go through unreferenced files and delete them.
    if delete:
        for name in unreferenced_files:
            if os.path.isfile(name) or os.path.islink(name):
                os.remove(name)
            elif os.path.isdir(name):
                shutil.rmtree(name)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from guardian.shortcuts import assign_perm
from rest_framework.test import APIRequestFactory

from resolwe.flow.models import Collection, Data, Entity, Process
from resolwe.flow.serializers import CollectionSerializer, EntitySerializer
from resolwe.test import TestCase


class CollectionSerializerTest(TestCase):
    def setUp(self):
        super(CollectionSerializerTest, self).setUp()

        process = Process.objects.create(
            name='Test process',
            contributor=self.contributor,
        )

        self.collection = Collection.objects.create(
            name='Test collection',
            contributor=self.contributor,
        )
        # In collection, user has `view` permission
        self.data_1 = Data.objects.create(
            name='First data',
            contributor=self.contributor,
            process=process,
        )
        # In collection, user has no permissions
        self.data_2 = Data.objects.create(
            name='Second data',
            contributor=self.contributor,
            process=process,
        )
        # Not in collection, user has `view` permission
        self.data_3 = Data.objects.create(
            name='Third data',
            contributor=self.contributor,
            process=process,
        )

        self.collection.data.add(self.data_1, self.data_2)

        assign_perm('view_collection', self.user, self.collection)
        assign_perm('view_data', self.user, self.data_1)
        assign_perm('view_data', self.user, self.data_3)

        self.factory = APIRequestFactory()

    def test_filter_data_permissions(self):
        request = self.factory.get('/')
        request.user = self.user
        request.query_params = {}

        serializer = CollectionSerializer(self.collection, context={'request': request})
        self.assertEqual(serializer.data['data'], [self.data_1.pk])


class EntitySerializerTest(TestCase):
    def setUp(self):
        super(EntitySerializerTest, self).setUp()

        process = Process.objects.create(
            name='Test process',
            contributor=self.contributor,
        )

        self.entity = Entity.objects.create(
            name='Test entity',
            contributor=self.contributor,
        )
        # In entity, user has `view` permission
        self.data_1 = Data.objects.create(
            name='First data',
            contributor=self.contributor,
            process=process,
        )
        # In entity, user has no permissions
        self.data_2 = Data.objects.create(
            name='Second data',
            contributor=self.contributor,
            process=process,
        )
        # Not in entity, user has `view` permission
        self.data_3 = Data.objects.create(
            name='Third data',
            contributor=self.contributor,
            process=process,
        )

        self.entity.data.add(self.data_1, self.data_2)

        assign_perm('view_entity', self.user, self.entity)
        assign_perm('view_data', self.user, self.data_1)
        assign_perm('view_data', self.user, self.data_3)

        self.factory = APIRequestFactory()

    def test_filter_data_permissions(self):
        request = self.factory.get('/')
        request.user = self.user
        request.query_params = {}

        serializer = EntitySerializer(self.entity, context={'request': request})
        self.assertEqual(serializer.data['data'], [self.data_1.pk])



""".. Ignore pydocstyle D400.

================
Flow Serializers
================

"""
from __future__ import absolute_import, division, print_function, unicode_literals

from django.contrib import auth
from django.contrib.auth.models import AnonymousUser
from django.db import transaction

from rest_framework import serializers, status
from rest_framework.exceptions import APIException, ValidationError
from rest_framework.fields import empty

from resolwe.flow.models import Collection, Data, DescriptorSchema, Entity, Process, Relation, RelationType, Storage
from resolwe.flow.models.entity import PositionInRelation
from resolwe.permissions.shortcuts import get_objects_for_user
from resolwe.rest.fields import ProjectableJSONField
from resolwe.rest.serializers import SelectiveFieldMixin


class NoContentError(APIException):
    """Content has not changed exception."""

    status_code = status.HTTP_204_NO_CONTENT
    detail = 'The content has not changed'


class ContributorSerializer(SelectiveFieldMixin, serializers.ModelSerializer):
    """Serializer for contributor User objects."""

    class Meta:
        """Serializer configuration."""

        # The model needs to be determined when instantiating the serializer class as
        # the applications may not yet be ready at this point.
        model = None
        fields = ('id', 'username', 'first_name', 'last_name')

    def __init__(self, instance=None, data=empty, **kwargs):
        """Initialize attributes."""
        # Use the correct User model.
        if self.Meta.model is None:
            self.Meta.model = auth.get_user_model()

        super(ContributorSerializer, self).__init__(instance, data, **kwargs)

    def to_internal_value(self, data):
        """Format the internal value."""
        # When setting the contributor, it may be passed as an integer.
        if isinstance(data, dict) and isinstance(data.get('id', None), int):
            data = data['id']
        elif isinstance(data, int):
            pass
        else:
            raise ValidationError("Contributor must be an integer or a dictionary with key 'id'")

        return self.Meta.model.objects.get(pk=data)


class ResolweBaseSerializer(SelectiveFieldMixin, serializers.ModelSerializer):
    """Base serializer for all `Resolwe` objects.

    This class is inherited from `django_rest_framework`'s
    `ModelSerialzer` class. The difference is that
    `update_protected_fields` are removed from `data` dict when update
    is performed.

    To check whether the class is called to create an instance or
    to update an existing one, it checks its value. If the value is
    `None`, a new instance is being created.
    The `update_protected_fields` tuple can be defined in the `Meta`
    class of child class.

    `NoContentError` is raised if no data would be changed, so we
    prevent changing `modified` field.

    """

    contributor = ContributorSerializer()

    def __init__(self, instance=None, data=empty, **kwargs):
        """Initialize attributes."""
        if (instance is not None and data is not empty and
                hasattr(self.Meta, 'update_protected_fields')):  # pylint: disable=no-member
            for field in self.Meta.update_protected_fields:  # pylint: disable=no-member
                if field in data:
                    data.pop(field)  # pylint: disable=no-member

            # prevent changing `modified` field if no field would be changed
            if set(data.keys()).issubset(set(self.Meta.read_only_fields)):  # pylint: disable=no-member
                raise NoContentError()

        super(ResolweBaseSerializer, self).__init__(instance, data, **kwargs)

    @property
    def request(self):
        """Extract request object from serializer context."""
        return self.context.get('request', None)


class ProcessSerializer(ResolweBaseSerializer):
    """Serializer for Process objects."""

    class Meta:
        """ProcessSerializer Meta options."""

        model = Process
        update_protected_fields = ('contributor', )
        read_only_fields = ('id', 'created', 'modified')
        fields = ('slug', 'name', 'data_name', 'version', 'type', 'flow_collection', 'category',
                  'persistence', 'description', 'input_schema', 'output_schema', 'requirements',
                  'run', 'scheduling_class') + update_protected_fields + read_only_fields


class DescriptorSchemaSerializer(ResolweBaseSerializer):
    """Serializer for DescriptorSchema objects."""

    schema = ProjectableJSONField(required=False)

    class Meta:
        """TemplateSerializer Meta options."""

        model = DescriptorSchema
        update_protected_fields = ('contributor', )
        read_only_fields = ('id', 'created', 'modified')
        fields = ('slug', 'name', 'version', 'schema') + update_protected_fields + read_only_fields


class DataSerializer(ResolweBaseSerializer):
    """Serializer for Data objects."""

    input = ProjectableJSONField(required=False)
    output = ProjectableJSONField(required=False)
    descriptor = ProjectableJSONField(required=False)
    process_slug = serializers.CharField(source='process.slug', read_only=True)
    process_name = serializers.CharField(source='process.name', read_only=True)
    process_type = serializers.CharField(source='process.type', read_only=True)
    process_input_schema = ProjectableJSONField(source='process.input_schema', read_only=True)
    process_output_schema = ProjectableJSONField(source='process.output_schema', read_only=True)

    name = serializers.CharField(read_only=False, required=False)
    slug = serializers.CharField(read_only=False, required=False)

    class Meta:
        """DataSerializer Meta options."""

        model = Data
        update_protected_fields = ('contributor', 'process',)
        read_only_fields = (
            'id', 'created', 'modified', 'started', 'finished', 'checksum', 'status', 'process_progress', 'process_rc',
            'process_info', 'process_warning', 'process_error', 'process_type', 'process_input_schema',
            'process_output_schema', 'process_slug', 'process_name', 'descriptor_dirty',
        )
        fields = (
            'slug', 'name', 'contributor', 'input', 'output', 'descriptor_schema', 'descriptor', 'tags',
            'process_memory', 'process_cores',
        ) + update_protected_fields + read_only_fields

    def get_fields(self):
        """Dynamically adapt fields based on the current request."""
        fields = super(DataSerializer, self).get_fields()

        if self.request.method == "GET":
            fields['descriptor_schema'] = DescriptorSchemaSerializer(required=False)
        else:
            fields['descriptor_schema'] = serializers.PrimaryKeyRelatedField(
                queryset=DescriptorSchema.objects.all(), required=False
            )

        return fields


class CollectionSerializer(ResolweBaseSerializer):
    """Serializer for Collection objects."""

    slug = serializers.CharField(read_only=False, required=False)
    settings = ProjectableJSONField(required=False)
    descriptor = ProjectableJSONField(required=False)

    class Meta:
        """CollectionSerializer Meta options."""

        model = Collection
        update_protected_fields = ('contributor',)
        read_only_fields = ('id', 'created', 'modified', 'descriptor_dirty')
        fields = ('slug', 'name', 'description', 'settings', 'descriptor_schema', 'descriptor',
                  'data') + update_protected_fields + read_only_fields

    def _serialize_data(self, data):
        """Return serialized data or list of ids, depending on `hydrate_data` query param."""
        if self.request and self.request.query_params.get('hydrate_data', False):
            serializer = DataSerializer(data, many=True, read_only=True)
            serializer.bind('data', self)
            return serializer.data
        else:
            return [d.id for d in data]

    def _filter_queryset(self, perms, queryset):
        """Filter object objects by permissions of user in request."""
        user = self.request.user if self.request else AnonymousUser()
        return get_objects_for_user(user, perms, queryset)

    def get_data(self, collection):
        """Return serialized list of data objects on collection that user has `view` permission on."""
        data = self._filter_queryset('view_data', collection.data.all())

        return self._serialize_data(data)

    def get_fields(self):
        """Dynamically adapt fields based on the current request."""
        fields = super(CollectionSerializer, self).get_fields()

        if self.request.method == "GET":
            fields['data'] = serializers.SerializerMethodField()
        else:
            fields['data'] = serializers.PrimaryKeyRelatedField(many=True, read_only=True)

        if self.request.method == "GET":
            fields['descriptor_schema'] = DescriptorSchemaSerializer(required=False)
        else:
            fields['descriptor_schema'] = serializers.PrimaryKeyRelatedField(
                queryset=DescriptorSchema.objects.all(), required=False
            )

        return fields


class EntitySerializer(CollectionSerializer):
    """Serializer for Entity."""

    collections = serializers.PrimaryKeyRelatedField(many=True, read_only=True)

    class Meta(CollectionSerializer.Meta):
        """Serializer configuration."""

        model = Entity
        fields = CollectionSerializer.Meta.fields + ('collections', 'descriptor_completed', 'tags')

    def get_data(self, entity):
        """Return serialized list of data objects on entity that user has `view` permission on."""
        data = self._filter_queryset('view_data', entity.data.all())

        return self._serialize_data(data)


class StorageSerializer(ResolweBaseSerializer):
    """Serializer for Storage objects."""

    json = ProjectableJSONField()

    class Meta:
        """StorageSerializer Meta options."""

        model = Storage
        update_protected_fields = ('contributor', )
        read_only_fields = ('id', 'created', 'modified')
        fields = ('slug', 'name', 'data', 'json') + update_protected_fields + read_only_fields


class PositionInRelationSerializer(SelectiveFieldMixin, serializers.ModelSerializer):
    """Serializer for PositionInRelation objects."""

    position = ProjectableJSONField(allow_null=True, required=False)

    class Meta:
        """PositionInRelationSerializer Meta options."""

        model = PositionInRelation
        fields = ('entity', 'position')


class RelationSerializer(ResolweBaseSerializer):
    """Serializer for Relation objects."""

    entities = PositionInRelationSerializer(source='positioninrelation_set', many=True)
    collection = serializers.PrimaryKeyRelatedField(queryset=Collection.objects.all(), required=True)

    class Meta:
        """RelationSerializer Meta options."""

        model = Relation
        update_protected_fields = ('contributor', 'entities', 'type')
        read_only_fields = ('id', 'created', 'modified')
        fields = ('collection', 'entities', 'label') + update_protected_fields + read_only_fields

    def get_fields(self):
        """Dynamically adapt fields based on the current request."""
        fields = super(RelationSerializer, self).get_fields()

        if self.request.method == "GET":
            fields['type'] = serializers.CharField(source='type.name')
        else:
            fields['type'] = serializers.PrimaryKeyRelatedField(queryset=RelationType.objects.all())

        return fields

    def create(self, validated_data):
        """Create ``Relation`` object and add ``Entities``."""
        # `entities` field is renamed to `positioninrelation_set` based on source of nested serializer

        entities = validated_data.pop('positioninrelation_set', {})

        # Prevent "Failed to populate slug Relation.slug from name" output
        validated_data['name'] = 'Relation'

        with transaction.atomic():
            instance = Relation.objects.create(**validated_data)
            for entity in entities:
                PositionInRelation.objects.create(
                    relation=instance,
                    entity=entity['entity'],
                    position=entity.get('position', None)
                )

        return instance



from __future__ import absolute_import, division, print_function, unicode_literals

# This will make sure the app is always imported when
# Django starts so that shared_task will use this app.
from .celery_conf import app as celery_app



# -*- coding: utf-8 -*-

import os

import django
from django.conf import settings

base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Manual setup is required for standalone Django usage
# NOTE: Since documentation is built using the built/installed package when
# using Tox, it can't use the 'test.settings' Django settings module.
settings.configure(
    INSTALLED_APPS=(
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'django.contrib.sessions',
        'django.contrib.staticfiles',
        'rest_framework',
        'guardian',
        'mathfilters',
        'versionfield',
        'resolwe',
        'resolwe.permissions',
        'resolwe.flow',
        'resolwe.elastic',
    ),
)
django.setup()

# Get package metadata from 'resolwe/__about__.py' file
about = {}
with open(os.path.join(base_dir, 'resolwe', '__about__.py')) as f:
    exec(f.read(), about)

# -- General configuration ------------------------------------------------

# The extension modules to enable.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
]

# The suffix(es) of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Resolwe'
version = about['__version__']
release = version
author = about['__author__']
copyright = about['__copyright__']

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Warn about all references where the target cannot be found
nitpicky = True
# Except for the following:
nitpick_ignore = [
    # This is needed to prevent warnings for container types, e.g.:
    # :type foo: tuple(bool, str)
    ('py:obj', ''),
]

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.
html_theme = 'sphinx_rtd_theme'

# Output file base name for HTML help builder.
htmlhelp_basename = 'Resolwedoc'

# Configuration for intersphinx
_django_major_version = "{}.{}".format(*django.VERSION[:2])
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'django': ('https://docs.djangoproject.com/en/{}/'.format(_django_major_version),
               'https://docs.djangoproject.com/en/{}/_objects/'.format(_django_major_version)),
}

# Configuration for extlinks
extlinks = {
    'drf': ('http://www.django-rest-framework.org/api-guide/%s', ''),
}



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-15 12:42
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0024_add_relations'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='collection',
            options={'default_permissions': (), 'get_latest_by': 'version', 'permissions': (('view_collection', 'Can view collection'), ('edit_collection', 'Can edit collection'), ('share_collection', 'Can share collection'), ('download_collection', 'Can download files from collection'), ('add_collection', 'Can add data objects to collection'), ('owner_collection', 'Is owner of the collection'))},
        ),
        migrations.AlterModelOptions(
            name='data',
            options={'default_permissions': (), 'get_latest_by': 'version', 'permissions': (('view_data', 'Can view data'), ('edit_data', 'Can edit data'), ('share_data', 'Can share data'), ('download_data', 'Can download files from data'), ('owner_data', 'Is owner of the data'))},
        ),
        migrations.AlterModelOptions(
            name='descriptorschema',
            options={'default_permissions': (), 'get_latest_by': 'version', 'permissions': (('view_descriptorschema', 'Can view descriptor schema'), ('edit_descriptorschema', 'Can edit descriptor schema'), ('share_descriptorschema', 'Can share descriptor schema'), ('owner_descriptorschema', 'Is owner of the description schema'))},
        ),
        migrations.AlterModelOptions(
            name='entity',
            options={'default_permissions': (), 'get_latest_by': 'version', 'permissions': (('view_entity', 'Can view entity'), ('edit_entity', 'Can edit entity'), ('share_entity', 'Can share entity'), ('download_entity', 'Can download files from entity'), ('add_entity', 'Can add data objects to entity'), ('owner_entity', 'Is owner of the entity'))},
        ),
        migrations.AlterModelOptions(
            name='process',
            options={'default_permissions': (), 'get_latest_by': 'version', 'permissions': (('view_process', 'Can view process'), ('share_process', 'Can share process'), ('owner_process', 'Is owner of the process'))},
        ),
        migrations.AlterModelOptions(
            name='storage',
            options={'default_permissions': (), 'get_latest_by': 'version'},
        ),
    ]



""".. Ignore pydocstyle D400.

==============
Flow Executors
==============

.. automodule:: resolwe.flow.executors.__main__
.. automodule:: resolwe.flow.executors.docker
.. automodule:: resolwe.flow.executors.local
.. automodule:: resolwe.flow.executors.null

"""



""".. Ignore pydocstyle D400.

===================================
Resolwe Test Helpers and Decorators
===================================

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import functools
import os
import shlex
import shutil
import subprocess
import unittest

import six
import wrapt

from django.conf import settings
from django.test import override_settings, tag

if six.PY2:
    # Monkey-patch shutil package with which function (available in Python 3.3+)
    import shutilwhich  # pylint: disable=import-error,unused-import

__all__ = (
    'check_installed', 'check_docker', 'with_custom_executor', 'with_docker_executor',
    'with_null_executor', 'with_resolwe_host', 'is_testing',
)

TAG_PROCESS = 'resolwe.process'


def check_installed(command):
    """Check if the given command is installed.

    :param str command: name of the command

    :return: (indicator of the availability of the command, message saying
              command is not available)
    :rtype: tuple(bool, str)

    """
    if shutil.which(command):
        return True, ""
    else:
        return False, "Command '{}' is not found.".format(command)


def check_docker():
    """Check if Docker is installed and working.

    :return: (indicator of the availability of Docker, reason for
              unavailability)
    :rtype: tuple(bool, str)

    """
    command = getattr(settings, 'FLOW_DOCKER_COMMAND', 'docker')
    info_command = '{} info'.format(command)
    available, reason = True, ""
    # TODO: Use subprocess.DEVNULL after dropping support for Python 2
    with open(os.devnull, 'wb') as DEVNULL:  # pylint: disable=invalid-name
        try:
            subprocess.check_call(shlex.split(info_command), stdout=DEVNULL, stderr=subprocess.STDOUT)
        except OSError:
            available, reason = False, "Docker command '{}' not found".format(command)
        except subprocess.CalledProcessError:
            available, reason = (False, "Docker command '{}' returned non-zero "
                                        "exit status".format(info_command))
    return available, reason


def with_custom_executor(wrapped=None, **custom_executor_settings):
    """Decorate unit test to run processes with a custom executor.

    :param dict custom_executor_settings: custom ``FLOW_EXECUTOR``
        settings with which you wish to override the current settings

    """
    if wrapped is None:
        return functools.partial(with_custom_executor, **custom_executor_settings)

    # pylint: disable=missing-docstring
    @wrapt.decorator
    def wrapper(wrapped_method, instance, args, kwargs):
        from resolwe.flow.managers import manager  # To prevent circular imports.

        executor_settings = settings.FLOW_EXECUTOR.copy()
        executor_settings.update(custom_executor_settings)

        try:
            with override_settings(FLOW_EXECUTOR=executor_settings):
                with manager.override_settings(FLOW_EXECUTOR=executor_settings):
                    # Re-run engine discovery as the settings have changed.
                    manager.discover_engines()

                    # Re-run the post_register_hook
                    manager.get_executor().post_register_hook(verbosity=0)

                    # Run the actual unit test method.
                    return wrapped_method(*args, **kwargs)
        finally:
            # Re-run engine discovery as the settings have changed.
            manager.discover_engines()

    return wrapper(wrapped)  # pylint: disable=no-value-for-parameter


def with_docker_executor(wrapped=None):
    """Decorate unit test to run processes with the Docker executor."""
    if wrapped is None:
        return functools.partial(with_docker_executor)

    # pylint: disable=missing-docstring
    @wrapt.decorator
    def wrapper(wrapped_method, instance, args, kwargs):
        return unittest.skipUnless(*check_docker())(
            with_custom_executor(
                NAME='resolwe.flow.executors.docker',
            )(wrapped_method)
        )(*args, **kwargs)

    return wrapper(wrapped)  # pylint: disable=no-value-for-parameter


@wrapt.decorator
def with_null_executor(wrapped_method, instance, args, kwargs):
    """Decorate unit test to run processes with the Null executor."""
    return with_custom_executor(NAME='resolwe.flow.executors.null')(wrapped_method)(*args, **kwargs)


@wrapt.decorator
def with_resolwe_host(wrapped_method, instance, args, kwargs):
    """Decorate unit test to give it access to a live Resolwe host.

    Set ``RESOLWE_HOST_URL`` setting to the address where the testing
    live Resolwe host listens to.

    .. note::

        This decorator must be used with a (sub)class of
        :class:`~django.test.LiveServerTestCase` which starts a live
        Django server in the background.

    """
    from resolwe.flow.managers import manager  # To prevent circular imports.

    if not hasattr(instance, 'server_thread'):
        raise AttributeError(
            "with_resolwe_host decorator must be used with a "
            "(sub)class of LiveServerTestCase that has the "
            "'server_thread' attribute"
        )
    resolwe_host_url = 'http://{host}:{port}'.format(
        host=instance.server_thread.host, port=instance.server_thread.port)
    with override_settings(RESOLWE_HOST_URL=resolwe_host_url):
        with manager.override_settings(RESOLWE_HOST_URL=resolwe_host_url):
            # Run the actual unit test method.
            return with_custom_executor(NETWORK='host')(wrapped_method)(*args, **kwargs)


def generate_process_tag(slug):
    """Generate test tag for a given process."""
    return '{}.{}'.format(TAG_PROCESS, slug)


def tag_process(*slugs):
    """Decorate unit test to tag it for a specific process."""
    slugs = [generate_process_tag(slug) for slug in slugs]
    slugs.append(TAG_PROCESS)  # Also tag with a general process tag.
    return tag(*slugs)


def has_process_tag(test, slug):
    """Check if a given test method/class is tagged for a specific process."""
    tags = getattr(test, 'tags', set())
    return generate_process_tag(slug) in tags


def get_processes_from_tags(test):
    """Extract process slugs from tags."""
    tags = getattr(test, 'tags', set())
    slugs = set()

    for tag_name in tags:
        if not tag_name.startswith('{}.'.format(TAG_PROCESS)):
            continue

        slugs.add(tag_name[len(TAG_PROCESS) + 1:])

    return slugs


def is_testing():
    """Return current testing status.

    This assumes that the Resolwe test runner is being used.
    """
    from resolwe.test_helpers.test_runner import is_testing  # pylint: disable=redefined-outer-name
    return is_testing()



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import copy

from django.conf import settings

from resolwe.flow.managers import manager
from resolwe.flow.models import Data, Process
from resolwe.test import TransactionTestCase


class EnvVarsTest(TransactionTestCase):

    def test_envvars(self):
        flow_executor = copy.copy(getattr(settings, 'FLOW_EXECUTOR', {}))
        flow_executor['SET_ENV'] = {
            'SET_ENV_TEST': 'test_var',
        }

        with manager.override_settings(FLOW_EXECUTOR=flow_executor, RESOLWE_HOST_URL='some.special.host'):
            process = Process.objects.create(
                name='Test environment variables',
                requirements={'expression-engine': 'jinja'},
                contributor=self.contributor,
                type='test:data:envvars:',
                input_schema=[],
                output_schema=[
                    {'name': 'resolweapihost', 'type': 'basic:string:'},
                    {'name': 'setenvtest', 'type': 'basic:string:'},
                ],
                run={
                    'language': 'bash',
                    'program': """
re-save resolweapihost $RESOLWE_HOST_URL
re-save setenvtest $SET_ENV_TEST
"""
                }
            )

            data = Data.objects.create(
                name='Data object',
                contributor=self.contributor,
                process=process,
                input={},
            )

            # update output
            data = Data.objects.get(pk=data.pk)

            self.assertEqual(data.output['resolweapihost'], 'some.special.host')
            self.assertEqual(data.output['setenvtest'], 'test_var')



# pylint: disable=missing-docstring
from resolwe.elastic.composer import composer
from resolwe.test import TestCase


class MyExtendableType(object):
    pass


class ComposerTest(TestCase):
    def test_add_extension(self):
        # Add some extensions.
        composer.add_extension('resolwe.elastic.tests.test_composer.MyExtendableType', 42)
        composer.add_extension('resolwe.elastic.tests.test_composer.MyExtendableType', 'foo')

        # Check that they are all here.
        extensions = composer.get_extensions(MyExtendableType)
        self.assertEqual(extensions, [42, 'foo'])

        extensions = composer.get_extensions('resolwe.elastic.tests.test_composer.MyExtendableType')
        self.assertEqual(extensions, [42, 'foo'])

        extensions = composer.get_extensions(MyExtendableType())
        self.assertEqual(extensions, [42, 'foo'])



""".. Ignore pydocstyle D400.

===============
Local Connector
===============

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import subprocess

from resolwe.utils import BraceMessage as __

from .base import BaseConnector

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Connector(BaseConnector):
    """Local connector for job execution."""

    def submit(self, data, runtime_dir, argv):
        """Run process locally.

        For details, see
        :meth:`~resolwe.flow.managers.workload_connectors.base.BaseConnector.submit`.
        """
        logger.debug(__(
            "Connector '{}' running for Data with id {} ({}).",
            self.__class__.__module__,
            data.id,
            repr(argv)
        ))
        subprocess.Popen(
            argv,
            cwd=runtime_dir,
            stdin=subprocess.DEVNULL
        ).wait()



"""Local workflow executor."""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from ..manager_commands import send_manager_command
from ..protocol import ExecutorProtocol
from ..run import BaseFlowExecutor

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class FlowExecutor(BaseFlowExecutor):  # pylint: disable=abstract-method
    """Null dataflow executor proxy.

    This executor is intended to be used in tests where you want to save
    the object to the database but don't need to run it.
    """

    name = 'null'

    def run(self, data_id, script):
        """Do nothing :)."""
        send_manager_command(ExecutorProtocol.FINISH)



"""Relation viewset."""
from __future__ import absolute_import, division, print_function, unicode_literals

import six

from django.db import IntegrityError, transaction

from rest_framework import exceptions, mixins, permissions, status, viewsets
from rest_framework.decorators import detail_route
from rest_framework.response import Response

from resolwe.flow.models import Relation
from resolwe.flow.models.entity import PositionInRelation, RelationType
from resolwe.flow.serializers import PositionInRelationSerializer, RelationSerializer
from resolwe.permissions.utils import assign_contributor_permissions


class RelationViewSet(mixins.CreateModelMixin,
                      mixins.ListModelMixin,
                      mixins.RetrieveModelMixin,
                      mixins.UpdateModelMixin,
                      mixins.DestroyModelMixin,
                      viewsets.GenericViewSet):
    """API view for :class:`Relation` objects."""

    queryset = Relation.objects.all().prefetch_related('contributor')
    serializer_class = RelationSerializer
    permission_classes = (permissions.IsAuthenticated,)
    ordering_fields = ('id', 'created', 'modified')
    ordering = ('id',)

    def _filter_queryset(self, queryset):
        """Filter queryset by queryparameters.

        Filtering is supported by ``name``, ``collection``, ``entities``
        and ``positions``.

        If ``positions`` parameter is given, it is combined with
        coresponding id in ``samples`` parameter and relations are
        filtered by (sample, position) pairs. Because of this, if
        ``positions`` is given, also ``samples`` must be given and
        they must be of the same length.

        NOTE: Because of complex filtering requirements it is not
              possible to use django_restframework_filters (at least
              not in a straight foreward way)
        """
        # query_params must be casted to dict, otherwise list values cannot be retrieved
        query_params = dict(self.request.query_params)

        id_ = query_params.get('id', None)
        relation_type = query_params.get('type', None)
        label = query_params.get('label', None)
        entities = query_params.get('entity', None)
        positions = query_params.get('position', None)
        collection = query_params.get('collection', None)

        if id_:
            queryset = queryset.filter(id=id_[0])

        if relation_type:
            queryset = queryset.filter(type__name=relation_type[0])

        if label:
            queryset = queryset.filter(label=label[0])

        if positions is not None and (entities is None or len(positions) != len(entities)):
            raise exceptions.ParseError(
                'If `positions` query parameter is given, also `entities` '
                'must be given and they must be of the same length.'
            )

        if collection:
            queryset = queryset.filter(collection__pk=collection[0])

        if entities:
            for entity, position in six.moves.zip_longest(entities, positions or []):
                filter_params = {'entities__pk': entity}
                if position:
                    filter_params['positioninrelation__position'] = position
                queryset = queryset.filter(**filter_params)

        return queryset

    def get_queryset(self):
        """Get queryset and perform custom filtering."""
        return self._filter_queryset(self.queryset)

    def create(self, request, *args, **kwargs):
        """Create a resource."""
        user = request.user
        if not user.is_authenticated:
            raise exceptions.NotFound

        relation_type = request.data.get('type')
        if not relation_type:
            return Response({'type': ['This field is required.']}, status=status.HTTP_400_BAD_REQUEST)

        rel_type_query = RelationType.objects.filter(name=relation_type)
        try:
            request.data['type'] = rel_type_query.last().pk
        except RelationType.DoesNotExist:
            return Response(
                {'type': ['Invalid type name "{}" - object does not exist.'.format(relation_type)]},
                status=status.HTTP_400_BAD_REQUEST)

        request.data['contributor'] = user.pk

        try:
            return super(RelationViewSet, self).create(request, *args, **kwargs)

        except IntegrityError as ex:
            return Response({u'error': str(ex)}, status=status.HTTP_409_CONFLICT)

    def perform_create(self, serializer):
        """Create a relation."""
        with transaction.atomic():
            instance = serializer.save()

            assign_contributor_permissions(instance)

    @detail_route(methods=[u'post'])
    def add_entity(self, request, pk=None):
        """Add ``Entity`` to ``Relation``."""
        relation = self.get_object()
        serializer = PositionInRelationSerializer(data=request.data, many=True)
        if not serializer.is_valid():
            Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)
        serializer.save(relation=relation)

        return Response()

    @detail_route(methods=[u'post'])
    def remove_entity(self, request, pk=None):
        """Remove data from collection."""
        if 'ids' not in request.data:
            return Response({"error": "`ids` parameter is required"}, status=status.HTTP_400_BAD_REQUEST)

        relation = self.get_object()
        for entity_id in request.data['ids']:
            PositionInRelation.objects.filter(relation=relation, entity=entity_id).delete()

        return Response()



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-28 10:23
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0021_entity_descriptor_completed'),
    ]

    operations = [
        migrations.AlterField(
            model_name='data',
            name='checksum',
            field=models.CharField(max_length=64, validators=[django.core.validators.RegexValidator(code='invalid_checksum', message='Checksum is exactly 40 alphanumerics', regex='^[0-9a-f]{64}$')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-07-15 06:49
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0008_fix_jsonfields'),
    ]

    operations = [
        migrations.AddField(
            model_name='data',
            name='parents',
            field=models.ManyToManyField(related_name='children', to='flow.Data'),
        ),
    ]



"""Central place for package metadata."""

# NOTE: We use __title__ instead of simply __name__ since the latter would
#       interfere with a global variable __name__ denoting object's name.
__title__ = 'resolwe'
__summary__ = 'Open source enterprise dataflow engine in Django'
__url__ = 'https://github.com/genialis/resolwe'

# Semantic versioning is used. For more information see:
# https://packaging.python.org/en/latest/distributing/#semantic-versioning-preferred
__version__ = '10.0.1'

__author__ = 'Genialis d.o.o.'
__email__ = 'dev-team@genialis.com'

__license__ = 'Apache License (2.0)'
__copyright__ = '2015-2018, ' + __author__

__all__ = (
    '__title__', '__summary__', '__url__', '__version__', '__author__',
    '__email__', '__license__', '__copyright__',
)



""".. Ignore pydocstyle D400.

===================================
Abstract Workload Manager Connector
===================================

"""


class BaseConnector(object):
    """The abstract base class for workload manager connectors.

    The main :class:`~resolwe.flow.managers.dispatcher.Manager` instance
    in :data:`~resolwe.flow.managers.manager` uses connectors to handle
    communication with concrete backend workload management systems,
    such as Celery and SLURM. The connectors need not worry about how
    jobs are discovered or how they're prepared for execution; this is
    all done by the manager.
    """

    def submit(self, data, runtime_dir, argv):
        """Submit the job to the workload management system.

        :param data: The :class:`~resolwe.flow.models.Data` object that
            is to be run.
        :param runtime_dir: The directory the executor is run from.
        :param argv: The argument vector used to spawn the executor.
        """
        raise NotImplementedError("Subclasses of BaseConnector must implement a submit() method.")



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-04-12 04:33
from __future__ import unicode_literals

import autoslug.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0003_support_sample'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('version',), max_length=100),
        ),
        migrations.AlterField(
            model_name='data',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('version',), max_length=100),
        ),
        migrations.AlterField(
            model_name='descriptorschema',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('version',), max_length=100),
        ),
        migrations.AlterField(
            model_name='process',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('version',), max_length=100),
        ),
        migrations.AlterField(
            model_name='storage',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('version',), max_length=100),
        ),
        migrations.AlterField(
            model_name='trigger',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('version',), max_length=100),
        ),
    ]



# pylint: disable=missing-docstring,invalid-name
from __future__ import absolute_import, division, print_function, unicode_literals

import six

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group
from django.contrib.contenttypes.models import ContentType
from django.db.models.query import QuerySet

from guardian.compat import get_user_permission_full_codename
from guardian.exceptions import MixedContentTypeError, WrongAppError
from guardian.shortcuts import assign_perm, get_objects_for_group, remove_perm
from rest_framework import status
from rest_framework.test import APIRequestFactory, force_authenticate

from resolwe.flow.models import Collection, Data, Process, Storage
from resolwe.flow.views import StorageViewSet
from resolwe.permissions.shortcuts import get_object_perms, get_objects_for_user, get_user_group_perms
from resolwe.test import TestCase

factory = APIRequestFactory()  # pylint: disable=invalid-name


class UserGroupTestCase(TestCase):

    def setUp(self):
        super(UserGroupTestCase, self).setUp()

        self.group1 = Group.objects.create(name="Test group 1")
        self.group2 = Group.objects.create(name="Test group 2")

        self.collection = Collection.objects.create(
            contributor=self.contributor,
            name="Test collection",
        )

        # This collection is here to make sure that other permissions
        # don't affect tested queries.
        collection2 = Collection.objects.create(
            contributor=self.contributor,
            name="Test collection 2",
        )
        assign_perm("view_collection", self.contributor, collection2)
        assign_perm("view_collection", self.group1, collection2)

    def test_user(self):
        assign_perm("view_collection", self.contributor, self.collection)
        assign_perm("edit_collection", self.contributor, self.collection)

        user_perms, group_perms = get_user_group_perms(self.contributor, self.collection)

        self.assertEqual(len(group_perms), 0)
        six.assertCountEqual(self, user_perms, ["view_collection", "edit_collection"])

    def test_user_in_group(self):
        self.group1.user_set.add(self.contributor)
        assign_perm("view_collection", self.group1, self.collection)
        assign_perm("edit_collection", self.group1, self.collection)

        user_perms, group_perms = get_user_group_perms(self.contributor, self.collection)
        self.assertEqual(len(group_perms), 1)
        six.assertCountEqual(self, group_perms[0][2], ["view_collection", "edit_collection"])
        self.assertEqual(len(user_perms), 0)

        assign_perm("view_collection", self.contributor, self.collection)

        user_perms, group_perms = get_user_group_perms(self.contributor, self.collection)
        self.assertEqual(len(group_perms), 1)
        six.assertCountEqual(self, group_perms[0][2], ["view_collection", "edit_collection"])
        self.assertEqual(len(user_perms), 1)
        six.assertCountEqual(self, user_perms, ["view_collection"])

    def test_user_in_multiple_groups(self):
        self.group1.user_set.add(self.contributor)
        self.group2.user_set.add(self.contributor)
        assign_perm("view_collection", self.group1, self.collection)
        assign_perm("edit_collection", self.group1, self.collection)
        assign_perm("view_collection", self.group2, self.collection)

        user_perms, group_perms = get_user_group_perms(self.contributor, self.collection)
        self.assertEqual(len(group_perms), 2)
        self.assertEqual(group_perms[0][0], self.group1.pk)
        six.assertCountEqual(self, group_perms[0][2], ["view_collection", "edit_collection"])
        self.assertEqual(group_perms[1][0], self.group2.pk)
        six.assertCountEqual(self, group_perms[1][2], ["view_collection"])
        self.assertEqual(len(user_perms), 0)

    def test_group(self):
        assign_perm("view_collection", self.group1, self.collection)
        assign_perm("edit_collection", self.group1, self.collection)
        user_perms, group_perms = get_user_group_perms(self.group1, self.collection)
        self.assertEqual(len(group_perms), 1)
        six.assertCountEqual(self, group_perms[0][2], ["view_collection", "edit_collection"])
        self.assertEqual(len(user_perms), 0)


class ObjectPermsTestCase(TestCase):

    def setUp(self):
        super(ObjectPermsTestCase, self).setUp()

        self.admin.delete()
        self.user1 = get_user_model().objects.create(username="test_user1")
        self.user2 = get_user_model().objects.create(username="test_user2")
        self.group1 = Group.objects.create(name="Test group 1")
        self.group2 = Group.objects.create(name="Test group 2")
        self.anonymous = AnonymousUser()

        self.collection = Collection.objects.create(
            contributor=self.user1,
            name="Test collection",
        )

    def _sort_perms(self, perms):
        for elm in perms:
            elm['permissions'] = sorted(elm['permissions'])
        return perms

    def test_all_permissions(self):
        self.group1.user_set.add(self.user1)

        perms = get_object_perms(self.collection)
        self.assertEqual(len(perms), 0)

        assign_perm("view_collection", self.user1, self.collection)
        assign_perm("edit_collection", self.user1, self.collection)
        assign_perm("view_collection", self.user2, self.collection)
        expected_perms = [
            {'permissions': ['edit', 'view'], 'type': 'user', 'id': self.user1.pk, 'name': 'test_user1'},
            {'permissions': ['view'], 'type': 'user', 'id': self.user2.pk, 'name': 'test_user2'},
        ]
        perms = get_object_perms(self.collection)
        six.assertCountEqual(self, self._sort_perms(expected_perms), self._sort_perms(perms))

        assign_perm("view_collection", self.group1, self.collection)
        assign_perm("edit_collection", self.group1, self.collection)
        assign_perm("view_collection", self.group2, self.collection)
        expected_perms.extend([
            {'permissions': ['edit', 'view'], 'type': 'group', 'id': self.group1.pk, 'name': 'Test group 1'},
            {'permissions': ['view'], 'type': 'group', 'id': self.group2.pk, 'name': 'Test group 2'},
        ])
        perms = get_object_perms(self.collection)
        six.assertCountEqual(self, self._sort_perms(expected_perms), self._sort_perms(perms))

        assign_perm("view_collection", self.anonymous, self.collection)
        expected_perms.append(
            {'permissions': ['view'], 'type': 'public'},
        )
        perms = get_object_perms(self.collection)
        six.assertCountEqual(self, self._sort_perms(expected_perms), self._sort_perms(perms))

    def test_user_permissions(self):
        self.group1.user_set.add(self.user1)
        assign_perm("view_collection", self.user1, self.collection)
        assign_perm("edit_collection", self.user1, self.collection)
        assign_perm("view_collection", self.user2, self.collection)
        assign_perm("view_collection", self.group1, self.collection)
        assign_perm("edit_collection", self.group1, self.collection)
        assign_perm("view_collection", self.group2, self.collection)

        expected_perms = [
            {'permissions': ['edit', 'view'], 'type': 'user', 'id': self.user1.pk, 'name': 'test_user1'},
            {'permissions': ['edit', 'view'], 'type': 'group', 'id': self.group1.pk, 'name': 'Test group 1'},
        ]
        perms = get_object_perms(self.collection, self.user1)

        six.assertCountEqual(self, self._sort_perms(expected_perms), self._sort_perms(perms))

        self.group2.user_set.add(self.user1)
        expected_perms.append(
            {'permissions': ['view'], 'type': 'group', 'id': self.group2.pk, 'name': 'Test group 2'},
        )
        perms = get_object_perms(self.collection, self.user1)
        six.assertCountEqual(self, self._sort_perms(expected_perms), self._sort_perms(perms))

        assign_perm("view_collection", self.anonymous, self.collection)
        expected_perms.append(
            {'permissions': ['view'], 'type': 'public'},
        )
        perms = get_object_perms(self.collection, self.user1)
        six.assertCountEqual(self, self._sort_perms(expected_perms), self._sort_perms(perms))


class StoragePermsTestCase(TestCase):

    def setUp(self):
        super(StoragePermsTestCase, self).setUp()

        proc = Process.objects.create(name='Test process', contributor=self.contributor)
        self.data = Data.objects.create(name='Test data', contributor=self.contributor, process=proc)
        dummy_data = Data.objects.create(name='Dummy data', contributor=self.contributor, process=proc)

        self.storage1 = Storage.objects.create(
            name='Test storage',
            json={},
            data=self.data,
            contributor=self.contributor,
        )

        self.storage2 = Storage.objects.create(
            name='Test storage 2',
            json={},
            data=self.data,
            contributor=self.contributor,
        )

        Storage.objects.create(
            name='Dummy storage',
            json={},
            data=dummy_data,
            contributor=self.contributor,
        )

        self.user = get_user_model().objects.create(username="test_user")
        self.group = Group.objects.create(name="test_group")

        self.storage_list_viewset = StorageViewSet.as_view(actions={
            'get': 'list',
        })

        self.storage_detail_viewset = StorageViewSet.as_view(actions={
            'get': 'retrieve',
        })

    def test_list_permissons(self):
        request = factory.get('/', content_type='application/json')
        force_authenticate(request, self.user)

        resp = self.storage_list_viewset(request)
        self.assertEqual(len(resp.data), 0)

        assign_perm("view_data", self.user, self.data)
        resp = self.storage_list_viewset(request)
        six.assertCountEqual(self, [storage['id'] for storage in resp.data],
                             [self.storage1.pk, self.storage2.pk])

        remove_perm("view_data", self.user, self.data)
        resp = self.storage_list_viewset(request)
        self.assertEqual(len(resp.data), 0)

    def test_detail_permissons(self):
        request = factory.get('/', content_type='application/json')
        force_authenticate(request, self.user)

        resp = self.storage_detail_viewset(request, pk=self.storage1.pk)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)

        assign_perm("view_data", self.user, self.data)
        resp = self.storage_detail_viewset(request, pk=self.storage1.pk)
        self.assertEqual(resp.data['name'], 'Test storage')

        remove_perm("view_data", self.user, self.data)
        resp = self.storage_detail_viewset(request, pk=self.storage1.pk)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)

    def test_not_allowed_methods(self):
        self.assertFalse(hasattr(StorageViewSet, 'update'))
        self.assertFalse(hasattr(StorageViewSet, 'partial_update'))
        self.assertFalse(hasattr(StorageViewSet, 'destroy'))
        self.assertFalse(hasattr(StorageViewSet, 'create'))


# tests copied from guardina.testapp.tests.test_shortcuts
class GetObjectsForUser(TestCase):

    def setUp(self):
        super(GetObjectsForUser, self).setUp()

        self.group = Group.objects.create(name='group')
        self.ctype = ContentType.objects.create(
            model='bar', app_label='fake-for-guardian-tests')

    def test_superuser(self):
        ctypes = ContentType.objects.all()
        objects = get_objects_for_user(self.admin,
                                       ['contenttypes.change_contenttype'], ctypes)
        self.assertEqual(set(ctypes), set(objects))

    def test_with_superuser_true(self):
        ctypes = ContentType.objects.all()
        objects = get_objects_for_user(self.admin,
                                       ['contenttypes.change_contenttype'], ctypes, with_superuser=True)
        self.assertEqual(set(ctypes), set(objects))

    def test_with_superuser_false(self):
        ctypes = ContentType.objects.all()
        obj1 = ContentType.objects.create(
            model='foo', app_label='guardian-tests')
        assign_perm('change_contenttype', self.admin, obj1)
        objects = get_objects_for_user(self.admin,
                                       ['contenttypes.change_contenttype'], ctypes, with_superuser=False)
        self.assertEqual(set([obj1]), set(objects))

    def test_anonymous(self):
        ctypes = ContentType.objects.all()
        objects = get_objects_for_user(self.contributor,
                                       ['contenttypes.change_contenttype'], ctypes)

        obj1 = ContentType.objects.create(
            model='foo', app_label='guardian-tests')
        assign_perm('change_contenttype', self.contributor, obj1)
        objects = get_objects_for_user(self.contributor,
                                       ['contenttypes.change_contenttype'], ctypes)
        self.assertEqual(set([obj1]), set(objects))

    def test_mixed_perms(self):
        codenames = [
            get_user_permission_full_codename('change'),
            'auth.change_permission',
        ]
        self.assertRaises(MixedContentTypeError, get_objects_for_user,
                          self.contributor, codenames)

    def test_perms_with_mixed_apps(self):
        codenames = [
            get_user_permission_full_codename('change'),
            'contenttypes.change_contenttype',
        ]
        self.assertRaises(MixedContentTypeError, get_objects_for_user,
                          self.contributor, codenames)

    def test_mixed_perms_and_klass(self):
        self.assertRaises(MixedContentTypeError, get_objects_for_user,
                          self.contributor, ['auth.change_group'], get_user_model())

    def test_no_app_label_nor_klass(self):
        self.assertRaises(WrongAppError, get_objects_for_user, self.contributor,
                          ['change_group'])

    def test_empty_perms_sequence(self):
        objects = get_objects_for_user(self.contributor, [], Group.objects.all())
        self.assertEqual(
            set(objects),
            set()
        )

    def test_perms_single(self):
        perm = 'auth.change_group'
        assign_perm(perm, self.contributor, self.group)
        self.assertEqual(
            set(get_objects_for_user(self.contributor, perm)),
            set(get_objects_for_user(self.contributor, [perm])))

    def test_klass_as_model(self):
        assign_perm('contenttypes.change_contenttype', self.contributor, self.ctype)

        objects = get_objects_for_user(self.contributor,
                                       ['contenttypes.change_contenttype'], ContentType)
        self.assertEqual([obj.name for obj in objects], [self.ctype.name])

    def test_klass_as_manager(self):
        assign_perm('auth.change_group', self.contributor, self.group)
        objects = get_objects_for_user(self.contributor, ['auth.change_group'],
                                       Group.objects)
        self.assertEqual([obj.name for obj in objects], [self.group.name])

    def test_klass_as_queryset(self):
        assign_perm('auth.change_group', self.contributor, self.group)
        objects = get_objects_for_user(self.contributor, ['auth.change_group'],
                                       Group.objects.all())
        self.assertEqual([obj.name for obj in objects], [self.group.name])

    def test_ensure_returns_queryset(self):
        objects = get_objects_for_user(self.contributor, ['auth.change_group'])
        self.assertTrue(isinstance(objects, QuerySet))

    def test_simple(self):
        group_names = ['group1', 'group2', 'group3']
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            assign_perm('change_group', self.contributor, group)

        objects = get_objects_for_user(self.contributor, ['auth.change_group'])
        self.assertEqual(len(objects), len(groups))
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(
            set(objects),
            set(groups))

    def test_multi_perms(self):
        group_names = ['group1', 'group2', 'group3']
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            assign_perm('auth.change_group', self.contributor, group)
        assign_perm('auth.delete_group', self.contributor, groups[1])

        objects = get_objects_for_user(self.contributor, ['auth.change_group',
                                                          'auth.delete_group'])
        self.assertEqual(len(objects), 1)
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(
            set(objects.values_list('name', flat=True)),
            set([groups[1].name]))

    def test_multi_perms_no_groups(self):
        group_names = ['group1', 'group2', 'group3']
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            assign_perm('auth.change_group', self.contributor, group)
        assign_perm('auth.delete_group', self.contributor, groups[1])

        objects = get_objects_for_user(self.contributor, ['auth.change_group',
                                                          'auth.delete_group'], use_groups=False)
        self.assertEqual(len(objects), 1)
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(
            set(objects.values_list('name', flat=True)),
            set([groups[1].name]))

    def test_any_of_multi_perms(self):
        group_names = ['group1', 'group2', 'group3']
        groups = [Group.objects.create(name=name) for name in group_names]
        assign_perm('auth.change_group', self.contributor, groups[0])
        assign_perm('auth.delete_group', self.contributor, groups[2])

        objects = get_objects_for_user(self.contributor, ['auth.change_group',
                                                          'auth.delete_group'], any_perm=True)
        self.assertEqual(len(objects), 2)
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(
            set(objects.values_list('name', flat=True)),
            set([groups[0].name, groups[2].name]))

    def test_groups_perms(self):
        group1 = Group.objects.create(name='group1')
        group2 = Group.objects.create(name='group2')
        group3 = Group.objects.create(name='group3')
        groups = [group1, group2, group3]
        for group in groups:
            self.contributor.groups.add(group)

        # Objects to operate on
        ctypes = list(ContentType.objects.all().order_by('id'))
        assign_perm('auth.change_group', self.contributor)
        assign_perm('change_contenttype', self.contributor, ctypes[0])
        assign_perm('change_contenttype', self.contributor, ctypes[1])
        assign_perm('delete_contenttype', self.contributor, ctypes[1])
        assign_perm('delete_contenttype', self.contributor, ctypes[2])

        assign_perm('change_contenttype', groups[0], ctypes[3])
        assign_perm('change_contenttype', groups[1], ctypes[3])
        assign_perm('change_contenttype', groups[2], ctypes[4])
        assign_perm('delete_contenttype', groups[0], ctypes[0])

        objects = get_objects_for_user(self.contributor,
                                       ['contenttypes.change_contenttype'])
        self.assertEqual(
            set(objects.values_list('id', flat=True)),
            set(ctypes[i].id for i in [0, 1, 3, 4]))

        objects = get_objects_for_user(self.contributor,
                                       ['contenttypes.change_contenttype',
                                        'contenttypes.delete_contenttype'])
        self.assertEqual(
            set(objects.values_list('id', flat=True)),
            set(ctypes[i].id for i in [0, 1]))

        objects = get_objects_for_user(self.contributor,
                                       ['contenttypes.change_contenttype'])
        self.assertEqual(
            set(objects.values_list('id', flat=True)),
            set(ctypes[i].id for i in [0, 1, 3, 4]))

    def test_has_global_permission_only(self):
        group_names = ['group1', 'group2', 'group3']
        for name in group_names:
            Group.objects.create(name=name)

        # global permission to change any group
        perm = 'auth.change_group'

        assign_perm(perm, self.contributor)
        objects = get_objects_for_user(self.contributor, perm)
        remove_perm(perm, self.contributor)
        self.assertEqual(set(objects),
                         set(Group.objects.all()))

    def test_has_global_permission_and_object_based_permission(self):
        group_names = ['group1', 'group2', 'group3']
        groups = [Group.objects.create(name=name) for name in group_names]
        # global permission to change any group
        perm_global = 'auth.change_group'
        perm_obj = 'delete_group'
        assign_perm(perm_global, self.contributor)
        assign_perm(perm_obj, self.contributor, groups[0])
        objects = get_objects_for_user(self.contributor, [perm_global, perm_obj])
        remove_perm(perm_global, self.contributor)
        self.assertEqual(set(objects.values_list('name', flat=True)),
                         set([groups[0].name]))

    def test_has_global_permission_and_object_based_permission_any_perm(self):
        group_names = ['group1', 'group2', 'group3']
        groups = [Group.objects.create(name=name) for name in group_names]
        # global permission to change any group
        perm_global = 'auth.change_group'
        # object based permission to change only a specific group
        perm_obj = 'auth.delete_group'
        assign_perm(perm_global, self.contributor)
        assign_perm(perm_obj, self.contributor, groups[0])
        objects = get_objects_for_user(
            self.contributor, [perm_global, perm_obj], any_perm=True, accept_global_perms=True)
        remove_perm(perm_global, self.contributor)
        self.assertEqual(set(objects),
                         set(Group.objects.all()))

    def test_object_based_permission_without_global_permission(self):
        group_names = ['group1', 'group2', 'group3']
        groups = [Group.objects.create(name=name) for name in group_names]
        # global permission to delete any group
        perm_global = 'auth.delete_group'
        perm_obj = 'auth.delete_group'
        assign_perm(perm_global, self.contributor)
        assign_perm(perm_obj, self.contributor, groups[0])
        objects = get_objects_for_user(
            self.contributor, [perm_obj], accept_global_perms=False)
        remove_perm(perm_global, self.contributor)
        self.assertEqual(set(objects.values_list('name', flat=True)),
                         set([groups[0].name]))

    def test_object_based_permission_with_groups_2perms(self):
        group_names = ['group1', 'group2', 'group3']
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            self.contributor.groups.add(group)
        # Objects to operate on
        ctypes = list(ContentType.objects.all().order_by('id'))
        assign_perm('contenttypes.change_contenttype', self.contributor)
        assign_perm('change_contenttype', self.contributor, ctypes[0])
        assign_perm('change_contenttype', self.contributor, ctypes[1])
        assign_perm('delete_contenttype', self.contributor, ctypes[1])
        assign_perm('delete_contenttype', self.contributor, ctypes[2])

        assign_perm('change_contenttype', groups[0], ctypes[3])
        assign_perm('change_contenttype', groups[1], ctypes[3])
        assign_perm('change_contenttype', groups[2], ctypes[4])
        assign_perm('delete_contenttype', groups[0], ctypes[0])

        objects = get_objects_for_user(self.contributor,
                                       ['contenttypes.change_contenttype',
                                        'contenttypes.delete_contenttype'], accept_global_perms=True)
        self.assertEqual(
            set(objects.values_list('id', flat=True)),
            set([ctypes[0].id, ctypes[1].id, ctypes[2].id]))

    def test_object_based_permission_with_groups_3perms(self):

        group_names = ['group1', 'group2', 'group3']
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            self.contributor.groups.add(group)
        # Objects to operate on
        ctypes = list(ContentType.objects.all().order_by('id'))
        assign_perm('contenttypes.change_contenttype', self.contributor)
        assign_perm('change_contenttype', self.contributor, ctypes[0])
        assign_perm('change_contenttype', self.contributor, ctypes[1])
        assign_perm('delete_contenttype', self.contributor, ctypes[1])
        assign_perm('delete_contenttype', self.contributor, ctypes[2])
        # add_contenttype does not make sense, here just for testing purposes,
        # to also cover one if branch in function.
        assign_perm('add_contenttype', self.contributor, ctypes[1])

        assign_perm('change_contenttype', groups[0], ctypes[3])
        assign_perm('change_contenttype', groups[1], ctypes[3])
        assign_perm('change_contenttype', groups[2], ctypes[4])
        assign_perm('delete_contenttype', groups[0], ctypes[0])
        assign_perm('add_contenttype', groups[0], ctypes[0])

        objects = get_objects_for_user(
            self.contributor, ['contenttypes.change_contenttype', 'contenttypes.delete_contenttype',
                               'contenttypes.add_contenttype'],
            accept_global_perms=True)
        self.assertEqual(
            set(objects.values_list('id', flat=True)),
            set([ctypes[0].id, ctypes[1].id]))

    def test_exception_different_ctypes(self):
        self.assertRaises(MixedContentTypeError, get_objects_for_user,
                          self.contributor, ['auth.change_permission', 'auth.change_group'])

    def test_has_any_permissions(self):
        # We use groups as objects.
        group_names = ['group1', 'group2', 'group3']
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            assign_perm('change_group', self.contributor, group)

        objects = get_objects_for_user(self.contributor, [], Group)
        self.assertEqual(len(objects), len(groups))
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(
            set(objects),
            set(groups))

    def test_short_codenames_with_klass(self):
        assign_perm('contenttypes.change_contenttype', self.contributor, self.ctype)

        objects = get_objects_for_user(self.contributor,
                                       ['change_contenttype'], ContentType)
        self.assertEqual([obj.name for obj in objects], [self.ctype.name])

    def test_has_any_group_permissions(self):
        # We use groups as objects.
        group_names = ['group1', 'group2', 'group3']
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            assign_perm('change_group', self.group, group)

        objects = get_objects_for_group(self.group, [], Group)
        self.assertEqual(len(objects), len(groups))
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(
            set(objects),
            set(groups))



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from resolwe.flow.models import Data, DescriptorSchema, Process
from resolwe.test import TestCase


class DescriptorTestCase(TestCase):
    def setUp(self):
        super(DescriptorTestCase, self).setUp()

        self.process = Process.objects.create(name="Dummy process", contributor=self.contributor)

        self.descriptor_schema = DescriptorSchema.objects.create(
            name='Descriptor schema',
            contributor=self.contributor,
            schema=[
                {'name': 'test_field', 'type': 'basic:string:', 'default': 'default value'}
            ]
        )

    def test_default_values(self):
        data = Data.objects.create(
            name='Data object',
            contributor=self.contributor,
            process=self.process,
            descriptor_schema=self.descriptor_schema,
        )
        self.assertEqual(data.descriptor['test_field'], 'default value')

        data = Data.objects.create(
            name='Data object 2',
            contributor=self.contributor,
            process=self.process,
            descriptor_schema=self.descriptor_schema,
            descriptor={'test_field': 'changed value'}
        )
        self.assertEqual(data.descriptor['test_field'], 'changed value')



"""Resolwe entity model."""
from __future__ import absolute_import, division, print_function, unicode_literals

from django.contrib.postgres.fields import ArrayField, JSONField
from django.db import models

from .base import BaseModel
from .collection import BaseCollection


class Entity(BaseCollection):
    """Postgres model for storing entities."""

    class Meta(BaseCollection.Meta):
        """Entity Meta options."""

        permissions = (
            ("view_entity", "Can view entity"),
            ("edit_entity", "Can edit entity"),
            ("share_entity", "Can share entity"),
            ("download_entity", "Can download files from entity"),
            ("add_entity", "Can add data objects to entity"),
            ("owner_entity", "Is owner of the entity"),
        )

    #: list of collections to which entity belongs
    collections = models.ManyToManyField('Collection')

    #: indicate whether `descriptor` is completed (set by user)
    descriptor_completed = models.BooleanField(default=False)

    #: tags for categorizing objects
    tags = ArrayField(models.CharField(max_length=255), default=list)


class RelationType(models.Model):
    """Model for storing relation types."""

    #: relation type name
    name = models.CharField(max_length=100, unique=True)

    #: indicates if order of entities in relation is important or not
    ordered = models.BooleanField(default=False)

    def __str__(self):
        """Format the object representation."""
        return self.name


class Relation(BaseModel):
    """Relations between entities."""

    class Meta(BaseCollection.Meta):
        """Relation Meta options."""

        permissions = (
            ("view_relation", "Can view relation"),
            ("edit_relation", "Can edit relation"),
            ("share_relation", "Can share relation"),
            ("owner_relation", "Is owner of the relation"),
        )

    #: type of the relation
    type = models.ForeignKey('RelationType', on_delete=models.PROTECT)

    #: collection in which relation is
    collection = models.ForeignKey('Collection', null=True, on_delete=models.SET_NULL)

    #: enteties in the relation
    entities = models.ManyToManyField(Entity, through='PositionInRelation')

    #: optional label of the relation (i.e. to further specify relation type)
    label = models.CharField(max_length=100, null=True, blank=True)


class PositionInRelation(models.Model):
    """Intermediate model to connect Entities with Relations."""

    #: entity in the relation
    entity = models.ForeignKey(Entity, on_delete=models.CASCADE)

    #: relation to which entity belongs
    relation = models.ForeignKey(Relation, on_delete=models.CASCADE)

    #: position of the entity in the relatin (relevant if `relation.type.odrered` is True)
    position = JSONField(null=True, blank=True)



"""Elastic search specialized dependencies for Resolwe."""
from __future__ import absolute_import, division, print_function, unicode_literals

from resolwe.elastic.builder import ManyToManyDependency
from resolwe.flow.models import Data


class DoneDataDependency(ManyToManyDependency):
    """Dependency only on DONE data objects."""

    def filter(self, obj, update_fields=None):
        """Determine if object should be processed."""
        return obj.status == Data.STATUS_DONE



""".. Ignore pydocstyle D400.

==================
List Docker images
==================

"""

import functools
import logging
import operator
import shlex
import subprocess
import threading

import yaml

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from resolwe.flow.executors.docker.constants import DEFAULT_CONTAINER_IMAGE
from resolwe.flow.models import Process

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

# Global list of images already pulled when running in the same process. This
# is used to avoid pulling images more than once when executed from tests.
PULLED_IMAGES = set()
PULLED_IMAGES_LOCK = threading.Lock()


class Command(BaseCommand):
    """List Docker images used by processes.  Optionally also pull them."""

    help = "List Docker images used by processes in either plain text or YAML (for Ansible)"

    def add_arguments(self, parser):
        """Add an argument to specify output format."""
        parser.add_argument('--format', dest='format', default='plain',
                            help="Set output format ('plain' [default] or 'yaml')")
        parser.add_argument('--pull', dest='pull', default=False, action='store_true',
                            help="Pull all images with Docker")
        parser.add_argument('--ignore-pull-errors', dest='ignore_pull_errors',
                            default=getattr(settings, 'FLOW_DOCKER_IGNORE_PULL_ERRORS', False),
                            action='store_true',
                            help="Don't fail whenever a Docker image can't be pulled")

    def handle(self, *args, **options):
        """Handle command list_docker_images."""
        verbosity = int(options.get('verbosity'))

        # Check that the specified output format is valid
        if options['format'] != 'plain' and options['format'] != 'yaml':
            raise CommandError("Unknown output format: %s" % options['format'])

        # Gather only unique latest custom Docker requirements that the processes are using
        # The 'image' field is optional, so be careful about that as well
        unique_docker_images = set(
            p.requirements['executor']['docker']['image']
            for p in Process.objects.order_by(
                'slug', '-version'
            ).distinct(
                'slug'
            ).only(
                'requirements'
            ).filter(
                requirements__icontains='docker'
            )
            if 'image' in p.requirements.get('executor', {}).get('docker', {})
        )

        # Add the default image.
        unique_docker_images.add(DEFAULT_CONTAINER_IMAGE)

        # Pull images if requested or just output the list in specified format
        if options['pull']:
            # Remove set of already pulled images.
            with PULLED_IMAGES_LOCK:
                unique_docker_images.difference_update(PULLED_IMAGES)

            # Get the desired 'docker' command from settings or use the default
            docker = getattr(settings, 'FLOW_DOCKER_COMMAND', 'docker')

            # Pull each image
            for img in unique_docker_images:
                ret = subprocess.call(
                    shlex.split('{} pull {}'.format(docker, img)),
                    stdout=None if verbosity > 0 else subprocess.DEVNULL,
                    stderr=None if verbosity > 0 else subprocess.DEVNULL,
                )

                # Update set of pulled images.
                with PULLED_IMAGES_LOCK:
                    PULLED_IMAGES.add(img)

                if ret != 0:
                    errmsg = "Failed to pull Docker image '{}'!".format(img)

                    if not options['ignore_pull_errors']:
                        # Print error and stop execution
                        raise CommandError(errmsg)
                    else:
                        # Print error, but keep going
                        logger.error(errmsg)
                        if verbosity > 0:
                            self.stderr.write(errmsg)
                else:
                    msg = "Docker image '{}' pulled successfully!".format(img)
                    logger.info(msg)
                    if verbosity > 0:
                        self.stdout.write(msg)
        else:
            # Sort the set of unique Docker images for nicer output.
            unique_docker_images = sorted(unique_docker_images)

            # Convert the set of unique Docker images into a list of dicts for easier output
            imgs = [
                dict(name=s[0], tag=s[1] if len(s) == 2 else 'latest')
                for s in (img.split(':') for img in unique_docker_images)
            ]

            # Output in YAML or plaintext (one image per line), as requested
            if options['format'] == 'yaml':
                out = yaml.safe_dump(imgs, default_flow_style=True, default_style="'")
            else:
                out = functools.reduce(operator.add,
                                       ('{name}:{tag}\n'.format(**i) for i in imgs), '')

            self.stdout.write(out, ending='')



"""Jinja2-based expression engine."""
from __future__ import absolute_import, division, print_function, unicode_literals

from contextlib import contextmanager
from importlib import import_module

import jinja2
from jinja2 import compiler
from jinja2.utils import soft_unicode
from six import string_types

from django.core.exceptions import ImproperlyConfigured

from resolwe.flow.expression_engines.base import BaseExpressionEngine
from resolwe.flow.expression_engines.exceptions import EvaluationError

from .filters import filters as builtin_filters


class NestedUndefined(jinja2.Undefined):
    """An undefined variable that can be nested."""

    def __unicode__(self):
        """Override."""
        return ''

    def __str__(self):
        """Override."""
        return ''

    def __call__(self, *args, **kwargs):
        """Override."""
        return NestedUndefined()

    def __getattr__(self, *args, **kwargs):
        """Override."""
        return NestedUndefined()

    def __getitem__(self, *args, **kwargs):
        """Override."""
        return NestedUndefined()


class SafeCodeGenerator(compiler.CodeGenerator):
    """Code generator wrapper."""

    def write_commons(self):
        """Override common preamble to inject our own escape wrappers."""
        super(SafeCodeGenerator, self).write_commons()

        self.writeline('escape = environment.escape')
        self.writeline('Markup = environment.markup_class')


class Environment(jinja2.Environment):
    """Custom Jinja2 environment."""

    # Custom code generator, which allows changing escape and markup.
    code_generator_class = SafeCodeGenerator

    # We usually don't output HTML so we ignore the markup class.
    markup_class = soft_unicode

    def __init__(self, engine):
        """Construct custom environment."""
        self._engine = engine

        super(Environment, self).__init__(
            undefined=NestedUndefined,
            autoescape=lambda _: self._engine._escape is not None  # pylint: disable=protected-access
        )

    def escape(self, value):
        """Escape given value."""
        value = soft_unicode(value)

        if self._engine._escape is None:  # pylint: disable=protected-access
            return value

        return self._engine._escape(value)  # pylint: disable=protected-access


class ExpressionEngine(BaseExpressionEngine):
    """Jinja2-based expression engine."""

    name = 'jinja'
    inline_tags = ('{{', '}}')

    def __init__(self, *args, **kwargs):
        """Construct the expression engine."""
        super(ExpressionEngine, self).__init__(*args, **kwargs)

        # Initialize the Jinja2 environment.
        # TODO: Should we use a sandboxed environment?
        self._environment = Environment(self)
        # Register built-in filters.
        self._environment.filters.update(builtin_filters)
        # Register custom filters.
        self._register_custom_filters()
        # Override the safe filter.
        self._environment.filters['safe'] = self._filter_mark_safe

        # Decorate all filters with our wrapper.
        for name, function in self._environment.filters.items():
            self._environment.filters[name] = self._wrap_jinja_filter(function)

        # Escape function and safe wrapper.
        self._escape = None
        self._safe_wrapper = None

    def _filter_mark_safe(self, value):
        """Filter to mark a value as safe."""
        if self._safe_wrapper is None:
            return value

        return self._safe_wrapper(value)

    def _wrap_jinja_filter(self, function):
        """Propagate exceptions as undefined values filter."""
        def wrapper(*args, **kwargs):
            """Filter wrapper."""
            try:
                return function(*args, **kwargs)
            except Exception:  # pylint: disable=broad-except
                return NestedUndefined()

        # Copy over Jinja filter decoration attributes.
        for attribute in dir(function):
            if attribute.endswith('filter'):
                setattr(wrapper, attribute, getattr(function, attribute))

        return wrapper

    def _register_custom_filters(self):
        """Register any custom filter modules."""
        custom_filters = self.settings.get('CUSTOM_FILTERS', [])
        if not isinstance(custom_filters, list):
            raise KeyError("`CUSTOM_FILTERS` setting must be a list.")

        for filter_module_name in custom_filters:
            try:
                filter_module = import_module(filter_module_name)
            except ImportError as error:
                raise ImproperlyConfigured(
                    "Failed to load custom filter module '{}'.\n"
                    "Error was: {}".format(filter_module_name, error)
                )

            try:
                filter_map = getattr(filter_module, 'filters')
                if not isinstance(filter_map, dict):
                    raise TypeError
            except (AttributeError, TypeError):
                raise ImproperlyConfigured(
                    "Filter module '{}' does not define a 'filters' dictionary".format(filter_module_name)
                )
            self._environment.filters.update(filter_map)

    @contextmanager
    def _evaluation_context(self, escape, safe_wrapper):
        """Configure the evaluation context."""
        self._escape = escape
        self._safe_wrapper = safe_wrapper

        try:
            yield
        finally:
            self._escape = None
            self._safe_wrapper = None

    def evaluate_block(self, template, context=None, escape=None, safe_wrapper=None):
        """Evaluate a template block."""
        if context is None:
            context = {}

        try:
            with self._evaluation_context(escape, safe_wrapper):
                template = self._environment.from_string(template)
                return template.render(**context)
        except jinja2.TemplateError as error:
            raise EvaluationError(error.args[0])
        finally:
            self._escape = None

    def evaluate_inline(self, expression, context=None, escape=None, safe_wrapper=None):
        """Evaluate an inline expression."""
        if context is None:
            context = {}

        try:
            with self._evaluation_context(escape, safe_wrapper):
                compiled = self._environment.compile_expression(expression)
                return compiled(**context)
        except jinja2.TemplateError as error:
            raise EvaluationError(error.args[0])



""".. Ignore pydocstyle D400.

=================================
Standalone Executor Contact Point
=================================

Command to run on local machine::

    ./manage.py runlistener

"""

import asyncio
from signal import SIGINT, SIGTERM, signal

from django.conf import settings
from django.core.management.base import BaseCommand

from resolwe.flow.managers.listener import ExecutorListener


class Command(BaseCommand):
    """Run the executor listener."""

    help = "Run the standalone manager contact point for executors."

    def add_arguments(self, parser):
        """Add command arguments."""
        super(Command, self).add_arguments(parser)
        parser.add_argument('--clear-queue', action='store_true',
                            help="Consume and ignore any outstanding messages in the result queue on startup.")

    def handle(self, *args, **kwargs):
        """Run the executor listener. This method never returns."""
        listener = ExecutorListener(redis_params=getattr(settings, 'FLOW_MANAGER', {}).get('REDIS_CONNECTION', {}))

        def _killer(signum, frame):
            """Kill the listener on receipt of a signal."""
            listener.terminate()
        signal(SIGINT, _killer)
        signal(SIGTERM, _killer)

        async def _runner():
            """Run the listener instance."""
            if kwargs['clear_queue']:
                await listener.clear_queue()
            async with listener:
                pass

        loop = asyncio.new_event_loop()
        loop.run_until_complete(_runner())
        loop.close()



""".. Ignore pydocstyle D400.

============
Celery Tasks
============

"""
from __future__ import absolute_import

import subprocess

from celery import shared_task  # pylint: disable=import-error


@shared_task
def celery_run(data_id, runtime_dir, argv):
    """Run process executor.

    :param data_id: The id of the :class:`~resolwe.flow.models.Data`
        object to be processed.
    :param runtime_dir: The directory from which to run the executor.
    :param argv: The argument vector used to run the executor.
    :param verbosity: The logging verbosity level.
    """
    subprocess.Popen(
        argv,
        cwd=runtime_dir,
        stdin=subprocess.DEVNULL
    ).wait()



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import shutil
import tempfile
import unittest

from mock import MagicMock, patch
from testfixtures import LogCapture

from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import override_settings
from django.utils.crypto import get_random_string

from resolwe.flow.managers.utils import disable_auto_calls
from resolwe.flow.models import Data, DescriptorSchema, Process
from resolwe.flow.utils import purge
from resolwe.test import ProcessTestCase


class PurgeTestFieldsMixin(object):
    """
    This class contains tests, which validate each field individually. It is used
    to perform basic unit tests (on simulated data, without running any processors)
    as well as more complex end-to-end tests (by running processors). Having them
    in this mixin makes sure that both, unit and e2e tests, are run for each field.
    """

    def test_basic_file(self):
        self.assertFieldWorks(
            'basic:file:',
            field_value={'file': 'but_not_this'},
            script_setup='touch remove_this but_not_this',
            script_save='re-save-file sample but_not_this',
            removed=['remove_this'],
            not_removed=['but_not_this'],
        )

    def test_basic_file_specialization(self):
        self.assertFieldWorks(
            'basic:file:html:',
            field_value={'file': 'but_not_this'},
            script_setup='touch remove_this but_not_this',
            script_save='re-save-file sample but_not_this',
            removed=['remove_this'],
            not_removed=['but_not_this'],
        )

    def test_basic_file_list(self):
        self.assertFieldWorks(
            'list:basic:file:',
            field_value=[{'file': 'but_not_this'}, {'file': 'and_not_this'}],
            script_setup='touch remove_this but_not_this and_not_this',
            script_save='re-save-file-list sample but_not_this and_not_this',
            removed=['remove_this'],
            not_removed=['but_not_this', 'and_not_this'],
        )

    def test_basic_dir(self):
        self.assertFieldWorks(
            'basic:dir:',
            field_value={'dir': 'but_not_this'},
            script_setup="""
mkdir remove_this but_not_this
touch remove_this/a remove_this/b remove_this/c
touch but_not_this/a but_not_this/b but_not_this/c
""",
            script_save='re-save-dir sample but_not_this',
            removed=['remove_this/', 'remove_this/a', 'remove_this/b', 'remove_this/c'],
            not_removed=['but_not_this/a', 'but_not_this/b', 'but_not_this/c'],
        )

    def test_basic_dir_list(self):
        self.assertFieldWorks(
            'list:basic:dir:',
            field_value=[{'dir': 'but_not_this'}, {'dir': 'and_not_this'}],
            script_setup="""
mkdir remove_this but_not_this and_not_this
touch remove_this/a remove_this/b remove_this/c
touch but_not_this/a but_not_this/b but_not_this/c
touch and_not_this/a and_not_this/b and_not_this/c
""",
            script_save='re-save-dir-list sample but_not_this and_not_this',
            removed=['remove_this/', 'remove_this/a', 'remove_this/b', 'remove_this/c'],
            not_removed=['but_not_this/a', 'but_not_this/b', 'but_not_this/c', 'and_not_this/a',
                         'and_not_this/b', 'and_not_this/c'],
        )


@override_settings(TEST_PROCESS_REQUIRE_TAGS=False)  # Test uses dynamic processes.
class PurgeE2ETest(PurgeTestFieldsMixin, ProcessTestCase):

    def create_and_run_processor(self, processor, **kwargs):
        processor_slug = get_random_string(6)
        Process.objects.create(
            slug=processor_slug,
            name='Test Purge Process',
            contributor=self.admin,
            type='data:test',
            version=1,
            **processor
        )

        return self.run_process(processor_slug, **kwargs)

    def assertFilesRemoved(self, data, *files):  # pylint: disable=invalid-name
        for name in files:
            path = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(data.pk), name)
            self.assertFalse(os.path.isfile(path), msg=path)

    def assertFilesNotRemoved(self, data, *files):  # pylint: disable=invalid-name
        for name in files:
            path = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(data.pk), name)
            self.assertTrue(os.path.isfile(path), msg=path)

    def test_complex_processor(self):
        data = self.create_and_run_processor(
            processor=dict(
                input_schema=[],
                output_schema=[
                    {
                        'name': 'sample_file',
                        'label': 'Sample output file',
                        'type': 'basic:file:'
                    },
                    {
                        'name': 'sample_dir',
                        'label': 'Sample output directory',
                        'type': 'basic:dir:'
                    },
                    {
                        'name': 'sample_file_list',
                        'label': 'Sample list of output files',
                        'type': 'list:basic:file:'
                    },
                    {
                        'name': 'sample_dir_list',
                        'label': 'Sample list of output directories',
                        'type': 'list:basic:dir:'
                    }
                ],
                run={
                    'language': 'bash',
                    'program': """
touch these files should be removed
touch this_file_should_stay
mkdir -p directory/should/be
touch directory/should/be/removed
touch directory/should/be/removed2
mkdir -p stay/directory
touch stay/directory/file1
touch stay/directory/file2
touch entry1 entry2 entry3
mkdir dir1 dir2 dir3
touch dir1/a dir2/b dir3/c dir3/d
touch ref1 ref2 ref3 ref4
mkdir refs
touch refs/a refs/b

re-save-file sample_file this_file_should_stay ref3
re-save-dir sample_dir stay ref4
re-save-file-list sample_file_list entry1 entry2:ref1 entry3:refs
re-save-dir-list sample_dir_list dir1:ref2 dir2 dir3
""",
                }
            )
        )

        self.assertFilesRemoved(data, 'these', 'files', 'should', 'be', 'removed', 'directory')
        self.assertFilesNotRemoved(data, 'this_file_should_stay', 'stay/directory/file1', 'stay/directory/file2',
                                   'entry1', 'entry2', 'entry3', 'dir1/a', 'dir2/b', 'dir3/c', 'dir3/d', 'ref1',
                                   'ref2', 'ref3', 'ref4', 'refs/a', 'refs/b')

    def assertFieldWorks(self, field_type, field_value, script_setup,  # pylint: disable=invalid-name
                         script_save, removed, not_removed):
        """
        Checks that a field is handled correctly when running a processor, which
        uses the field.
        """

        field_schema = {
            'name': 'sample',
            'label': 'Sample output',
            'type': field_type
        }

        # Test output.
        data = self.create_and_run_processor(
            processor=dict(
                input_schema=[],
                output_schema=[field_schema],
                run={
                    'language': 'bash',
                    'program': script_setup + '\n' + script_save
                }
            )
        )

        self.assertFilesRemoved(data, *removed)
        self.assertFilesNotRemoved(data, *not_removed)

        # Test descriptor.
        descriptor_schema = DescriptorSchema.objects.create(
            slug=get_random_string(6),
            contributor=self.admin,
            schema=[field_schema]
        )
        data = self.create_and_run_processor(
            processor=dict(
                input_schema=[],
                output_schema=[],
                run={
                    'language': 'bash',
                    'program': script_setup
                }
            ),
            descriptor_schema=descriptor_schema,
            descriptor={'sample': field_value}
        )

        self.assertFilesRemoved(data, *removed)
        self.assertFilesNotRemoved(data, *not_removed)

    @unittest.skipIf(True, "since PR308: manager now separated into parts, executor not logging here anymore")
    def test_exception_logging(self):
        with patch('resolwe.flow.utils.purge.os', wraps=os) as os_mock:
            # Ensure that purge raises an exception, so we can check whether the exception
            # gets logged correctly.
            class TestPurgeException(Exception):
                pass

            def exception_raiser(*args, **kwargs):
                raise TestPurgeException
            os_mock.walk = exception_raiser

            with LogCapture() as log:
                self.create_and_run_processor(
                    processor=dict(
                        input_schema=[],
                        output_schema=[{
                            'name': 'sample',
                            'label': 'Sample output',
                            'type': 'basic:file:'
                        }],
                        run={
                            'language': 'bash',
                            'program': 'touch sample && re-save-file sample sample'
                        }
                    )
                )

                self.assertEqual(len(log.records), 1)
                self.assertEqual(log.records[0].name, 'resolwe.flow.executors')
                self.assertTrue(str(log.records[0].msg).startswith('Purge error:'))
                self.assertTrue('TestPurgeException' in str(log.records[0].msg))


class PurgeUnitTest(PurgeTestFieldsMixin, ProcessTestCase):

    def assertFieldWorks(self, field_type, field_value, script_setup,  # pylint: disable=invalid-name
                         script_save, removed, not_removed):
        """
        Checks that a field is handled correctly by `get_purge_files` under a
        simulated Data object.
        """

        field_schema = {
            'name': 'sample',
            'label': 'Sample output',
            'type': field_type
        }

        # Test simulated operation.
        simulated_root = tempfile.mkdtemp()
        try:
            for filename in removed + not_removed:
                directory, basename = os.path.split(filename)
                if directory:
                    try:
                        os.makedirs(os.path.join(simulated_root, directory))
                    except OSError:
                        pass

                if basename:
                    with open(os.path.join(simulated_root, filename), 'w'):
                        pass

            unreferenced_files = purge.get_purge_files(
                simulated_root,
                output={'sample': field_value},
                output_schema=[field_schema],
                descriptor={},
                descriptor_schema=[]
            )

            def strip_slash(filename):
                return filename[:-1] if filename[-1] == '/' else filename

            for filename in not_removed:
                self.assertNotIn(strip_slash(os.path.join(simulated_root, filename)), unreferenced_files)

            for filename in removed:
                filename = strip_slash(os.path.join(simulated_root, filename))
                self.assertIn(filename, unreferenced_files)
                unreferenced_files.discard(filename)

            # Ensure that nothing more is removed.
            self.assertEqual(len(unreferenced_files), 0)
        finally:
            shutil.rmtree(simulated_root)

    def create_test_file(self, data, filename):
        """
        Creates an empty file in the proper directory, so that a file can be used
        as a reference in the Data object's output/descriptor.
        """

        if isinstance(data, Data):
            data_id = data.pk
        else:
            data_id = int(data)

        root = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(data_id))
        try:
            os.makedirs(root)
            self.test_files.add(root)
        except OSError:
            pass

        filename = os.path.join(root, filename)
        with open(filename, 'w'):
            self.test_files.add(filename)

    def setUp(self):
        super(PurgeUnitTest, self).setUp()

        self.test_files = set()

    def tearDown(self):
        for filename in self.test_files:
            if os.path.isfile(filename):
                os.remove(filename)
            elif os.path.isdir(filename):
                shutil.rmtree(filename)

        super(PurgeUnitTest, self).tearDown()

    # This patch is required so that the manager is not invoked while saving Data.
    @disable_auto_calls()
    def test_remove(self):
        user = get_user_model().objects.create(username="test_user")
        processor = Process.objects.create(
            name='Test process',
            contributor=user,
            output_schema=[
                {'name': 'sample', 'type': 'basic:file:'}
            ]
        )

        data = {
            'name': 'Test data',
            'contributor': user,
            'process': processor,
        }

        completed_data = Data.objects.create(**data)
        completed_data.status = Data.STATUS_DONE
        completed_data.output = {'sample': {'file': 'test-file'}}
        self.create_test_file(completed_data, 'test-file')
        self.create_test_file(completed_data, 'removeme')
        completed_data.save()

        pending_data = Data.objects.create(**data)
        self.create_test_file(pending_data, 'test-file')
        self.create_test_file(pending_data, 'donotremoveme')

        # Check that nothing is removed if delete is False (the default).
        with patch('resolwe.flow.utils.purge.os', wraps=os) as os_mock:
            os_mock.path.isfile = MagicMock(return_value=True)
            os_mock.remove = MagicMock()
            purge.data_purge()
            os_mock.remove.assert_not_called()

        # Check that only the 'removeme' file from the completed Data objects is removed
        # and files from the second (not completed) Data objects are unchanged.
        with patch('resolwe.flow.utils.purge.os', wraps=os) as os_mock:
            os_mock.path.isfile = MagicMock(return_value=True)
            os_mock.remove = MagicMock()
            purge.data_purge(delete=True)
            os_mock.remove.assert_called_once_with(
                os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(completed_data.pk), 'removeme'))

        # Create dummy data directories for non-existant data objects.
        self.create_test_file(990, 'dummy')
        self.create_test_file(991, 'dummy')

        # Check that only the 'removeme' file from the completed Data objects is removed
        # together with directories not belonging to any data objects.
        with patch('resolwe.flow.utils.purge.os', wraps=os) as os_mock:
            os_mock.path.isfile = MagicMock(return_value=True)
            os_mock.remove = MagicMock()
            purge.data_purge(delete=True)
            self.assertEqual(os_mock.remove.call_count, 3)
            os_mock.remove.assert_any_call(
                os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(completed_data.pk), 'removeme'))
            os_mock.remove.assert_any_call(
                os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], '990'))
            os_mock.remove.assert_any_call(
                os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], '991'))

        # Create another data object and check that if remove is called on one object,
        # only that object's data is removed.
        another_data = Data.objects.create(**data)
        another_data.status = Data.STATUS_DONE
        another_data.output = {'sample': {'file': 'test-file'}}
        self.create_test_file(another_data, 'test-file')
        self.create_test_file(another_data, 'removeme')
        another_data.save()

        with patch('resolwe.flow.utils.purge.os', wraps=os) as os_mock:
            os_mock.path.isfile = MagicMock(return_value=True)
            os_mock.remove = MagicMock()
            purge.data_purge(data_ids=[another_data.pk], delete=True)
            os_mock.remove.assert_called_once_with(
                os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(another_data.pk), 'removeme'))



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from django.db import models


class TestModel(models.Model):

    class Meta:
        permissions = (
            ('view_testmodel', "Can view model"),
            ('edit_testmodel', "Can edit model"),
        )

    name = models.CharField(max_length=30)

    field_process_type = models.CharField(max_length=100)

    number = models.IntegerField()

    date = models.DateTimeField(null=True)


class TestModelWithDependency(models.Model):

    name = models.CharField(max_length=30)

    dependencies = models.ManyToManyField('TestDependency')


class TestDependency(models.Model):

    name = models.CharField(max_length=30)


class TestSelfDependency(models.Model):

    name = models.CharField(max_length=30)

    dependencies = models.ManyToManyField('self', symmetrical=False, related_name='parents')



"""Base model for all Resolwe models."""
from __future__ import absolute_import, division, print_function, unicode_literals

from versionfield import VersionField

from django.conf import settings
from django.db import IntegrityError, models, transaction

from .fields import ResolweSlugField

VERSION_NUMBER_BITS = (8, 10, 14)

# Maximum number of slug-generation retries.
MAX_SLUG_RETRIES = 10


class BaseModel(models.Model):
    """Abstract model that includes common fields for other models."""

    class Meta:
        """BaseModel Meta options."""

        abstract = True
        unique_together = ('slug', 'version')
        default_permissions = ()
        get_latest_by = 'version'

    #: URL slug
    slug = ResolweSlugField(populate_from='name', unique_with='version', max_length=100)

    #: process version
    version = VersionField(number_bits=VERSION_NUMBER_BITS, default='0.0.0')

    #: object name
    name = models.CharField(max_length=100)

    #: creation date and time
    created = models.DateTimeField(auto_now_add=True, db_index=True)

    #: modified date and time
    modified = models.DateTimeField(auto_now=True, db_index=True)

    #: user that created the entry
    contributor = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)

    def __str__(self):
        """Format model name."""
        return self.name

    def save(self, *args, **kwargs):
        """Save the model."""
        for _ in range(MAX_SLUG_RETRIES):
            try:
                # Attempt to save the model. It may fail due to slug conflict.
                with transaction.atomic():
                    super(BaseModel, self).save(*args, **kwargs)
                    break
            except IntegrityError as error:
                # Retry in case of slug conflicts.
                if '{}_slug'.format(self._meta.db_table) in error.args[0]:
                    self.slug = None
                    continue

                raise
        else:
            raise IntegrityError("Maximum number of retries exceeded during slug generation")



""".. Ignore pydocstyle D400.

==================
Register Processes
==================

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import re

import jsonschema
import yaml
from versionfield.utils import convert_version_string_to_int

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand
from django.db.models import Max

from resolwe.flow.engine import InvalidEngineError
from resolwe.flow.finders import get_finders
from resolwe.flow.managers import manager
from resolwe.flow.models import DescriptorSchema, Process
from resolwe.flow.models.base import VERSION_NUMBER_BITS
from resolwe.flow.models.utils import validate_schema, validation_schema
from resolwe.flow.utils import dict_dot, iterate_schema
from resolwe.permissions.utils import assign_contributor_permissions, copy_permissions

PROCESSOR_SCHEMA = validation_schema('processor')
DESCRIPTOR_SCHEMA = validation_schema('descriptor')


class Command(BaseCommand):
    """Register processes."""

    help = 'Register processes'

    def add_arguments(self, parser):
        """Command arguments."""
        parser.add_argument('-s', '--schemas', type=str, nargs='*', help="process names to register")
        parser.add_argument('-f', '--force', action='store_true', help="register also if version mismatch")
        parser.add_argument('--path', type=str, nargs='*', default=[],
                            help="search paths for processes and descriptors")

    def valid(self, instance, schema):
        """Validate schema."""
        try:
            jsonschema.validate(instance, schema)
        except jsonschema.exceptions.ValidationError as ex:
            self.stderr.write("    VALIDATION ERROR: {}".format(instance['name'] if 'name' in instance else ''))
            self.stderr.write("        path:       {}".format(ex.path))
            self.stderr.write("        message:    {}".format(ex.message))
            self.stderr.write("        validator:  {}".format(ex.validator))
            self.stderr.write("        val. value: {}".format(ex.validator_value))
            return False

        try:
            # Check that default values fit field schema.
            for field in ['input', 'output', 'schema']:
                for schema, _, path in iterate_schema({}, instance.get(field, {})):
                    if 'default' in schema:
                        validate_schema({schema['name']: schema['default']}, [schema])
        except ValidationError:
            self.stderr.write("    VALIDATION ERROR: {}".format(instance['name']))
            self.stderr.write("        Default value of field '{}' is not valid.". format(path))
            return False

        return True

    def find_schemas(self, schema_path, filters=None, schema_type='process', verbosity=1):
        """Find schemas in packages that match filters."""
        schema_matches = []

        if not os.path.isdir(schema_path):
            if verbosity > 0:
                self.stdout.write("Invalid path {}".format(schema_path))
            return

        if schema_type not in ['process', 'descriptor']:
            raise ValueError('Invalid schema type')

        for root, _, files in os.walk(schema_path):
            for schema_file in [os.path.join(root, fn) for fn in files]:

                if not schema_file.lower().endswith(('.yml', '.yaml')):
                    continue

                with open(schema_file) as fn:
                    schemas = yaml.load(fn)
                if not schemas:
                    self.stderr.write("Could not read YAML file {}".format(schema_file))
                    continue

                for schema in schemas:
                    if filters and not (schema.get('slug', '') in filters or schema.get('name', '') in filters):
                        continue

                    if schema_type == 'process' and 'run' not in schema:
                        continue

                    if schema_type == 'descriptor' and 'schema' not in schema:
                        continue

                    schema_matches.append(schema)

        return schema_matches

    def register_processes(self, process_schemas, user, force=False, verbosity=1):
        """Read and register processors."""
        log_processors = []
        log_templates = []

        for p in process_schemas:
            if p['type'][-1] != ':':
                p['type'] += ':'

            if 'category' in p and not p['category'].endswith(':'):
                p['category'] += ':'

            for field in ['input', 'output']:
                for schema, _, _ in iterate_schema({}, p[field] if field in p else {}):
                    if not schema['type'][-1].endswith(':'):
                        schema['type'] += ':'
            # TODO: Check if schemas validate with our JSON meta schema and Processor model docs.

            if not self.valid(p, PROCESSOR_SCHEMA):
                continue

            if 'persistence' in p:
                persistence_mapping = {
                    'RAW': Process.PERSISTENCE_RAW,
                    'CACHED': Process.PERSISTENCE_CACHED,
                    'TEMP': Process.PERSISTENCE_TEMP,
                }

                p['persistence'] = persistence_mapping[p['persistence']]

            if 'scheduling_class' in p:
                scheduling_class_mapping = {
                    'interactive': Process.SCHEDULING_CLASS_INTERACTIVE,
                    'batch': Process.SCHEDULING_CLASS_BATCH
                }

                p['scheduling_class'] = scheduling_class_mapping[p['scheduling_class']]

            if 'input' in p:
                p['input_schema'] = p.pop('input')

            if 'output' in p:
                p['output_schema'] = p.pop('output')

            slug = p['slug']

            if 'run' in p:
                # Set default language to 'bash' if not set.
                p['run'].setdefault('language', 'bash')

                # Transform output schema using the execution engine.
                try:
                    execution_engine = manager.get_execution_engine(p['run']['language'])
                    extra_output_schema = execution_engine.get_output_schema(p)
                    if extra_output_schema:
                        p.setdefault('output_schema', []).extend(extra_output_schema)
                except InvalidEngineError:
                    self.stderr.write("Skip processor {}: execution engine '{}' not supported".format(
                        slug, p['run']['language']
                    ))
                    continue

            # Validate if container image is allowed based on the configured pattern.
            # NOTE: This validation happens here and is not deferred to executors because the idea
            #       is that this will be moved to a "container" requirement independent of the
            #       executor.
            if hasattr(settings, 'FLOW_CONTAINER_VALIDATE_IMAGE'):
                try:
                    container_image = dict_dot(p, 'requirements.executor.docker.image')
                    if not re.match(settings.FLOW_CONTAINER_VALIDATE_IMAGE, container_image):
                        self.stderr.write("Skip processor {}: container image does not match '{}'".format(
                            slug, settings.FLOW_CONTAINER_VALIDATE_IMAGE,
                        ))
                        continue
                except KeyError:
                    pass

            version = p['version']
            int_version = convert_version_string_to_int(version, VERSION_NUMBER_BITS)

            # `latest version` is returned as `int` so it has to be compared to `int_version`
            latest_version = Process.objects.filter(slug=slug).aggregate(Max('version'))['version__max']
            if latest_version is not None and latest_version > int_version:
                self.stderr.write("Skip processor {}: newer version installed".format(slug))
                continue

            previous_process_qs = Process.objects.filter(slug=slug)
            if previous_process_qs.exists():
                previous_process = previous_process_qs.latest()
            else:
                previous_process = None

            process_query = Process.objects.filter(slug=slug, version=version)
            if process_query.exists():
                if not force:
                    if verbosity > 0:
                        self.stdout.write("Skip processor {}: same version installed".format(slug))
                    continue

                process_query.update(**p)
                log_processors.append("Updated {}".format(slug))
            else:
                process = Process.objects.create(contributor=user, **p)
                assign_contributor_permissions(process)
                if previous_process:
                    copy_permissions(previous_process, process)
                log_processors.append("Inserted {}".format(slug))

        if verbosity > 0:
            if log_processors:
                self.stdout.write("Processor Updates:")
                for log in log_processors:
                    self.stdout.write("  {}".format(log))

            if log_templates:
                self.stdout.write("Default Template Updates:")
                for log in log_templates:
                    self.stdout.write("  {}".format(log))

    def register_descriptors(self, descriptor_schemas, user, force=False, verbosity=1):
        """Read and register descriptors."""
        log_descriptors = []

        for descriptor_schema in descriptor_schemas:
            for schema, _, _ in iterate_schema({}, descriptor_schema.get('schema', {})):
                if not schema['type'][-1].endswith(':'):
                    schema['type'] += ':'

            if 'schema' not in descriptor_schema:
                descriptor_schema['schema'] = []

            if not self.valid(descriptor_schema, DESCRIPTOR_SCHEMA):
                continue

            slug = descriptor_schema['slug']
            version = descriptor_schema.get('version', '0.0.0')
            int_version = convert_version_string_to_int(version, VERSION_NUMBER_BITS)

            # `latest version` is returned as `int` so it has to be compared to `int_version`
            latest_version = DescriptorSchema.objects.filter(slug=slug).aggregate(Max('version'))['version__max']
            if latest_version is not None and latest_version > int_version:
                self.stderr.write("Skip descriptor schema {}: newer version installed".format(slug))
                continue

            previous_descriptor_qs = DescriptorSchema.objects.filter(slug=slug)
            if previous_descriptor_qs.exists():
                previous_descriptor = previous_descriptor_qs.latest()
            else:
                previous_descriptor = None

            descriptor_query = DescriptorSchema.objects.filter(slug=slug, version=version)
            if descriptor_query.exists():
                if not force:
                    if verbosity > 0:
                        self.stdout.write("Skip descriptor schema {}: same version installed".format(slug))
                    continue

                descriptor_query.update(**descriptor_schema)
                log_descriptors.append("Updated {}".format(slug))
            else:
                descriptor = DescriptorSchema.objects.create(contributor=user, **descriptor_schema)
                assign_contributor_permissions(descriptor)
                if previous_descriptor:
                    copy_permissions(previous_descriptor, descriptor)
                log_descriptors.append("Inserted {}".format(slug))

        if log_descriptors and verbosity > 0:
            self.stdout.write("Descriptor schemas Updates:")
            for log in log_descriptors:
                self.stdout.write("  {}".format(log))

    def handle(self, *args, **options):
        """Register processes."""
        schemas = options.get('schemas')
        force = options.get('force')
        paths = options.get('path')

        verbosity = int(options.get('verbosity'))

        if not isinstance(paths, list):
            raise ValueError("Argument paths must be of type list")

        if schemas is not None and not isinstance(schemas, list):
            raise ValueError("Argument schemas must be of type list or None")

        users = get_user_model().objects.filter(is_superuser=True).order_by('date_joined')

        if not users.exists():
            self.stderr.write("Admin does not exist: create a superuser")
            exit(1)

        user_admin = users.first()

        processes_paths = paths[:]
        descriptors_paths = paths[:]

        if not paths:
            for finder in get_finders():
                processes_paths.extend(finder.find_processes())
                descriptors_paths.extend(finder.find_descriptors())

        process_schemas = []
        for proc_path in processes_paths:
            process_schemas.extend(
                self.find_schemas(proc_path, filters=schemas, schema_type='process', verbosity=verbosity))

        self.register_processes(process_schemas, user_admin, force, verbosity=verbosity)

        descriptor_schemas = []
        for desc_path in descriptors_paths:
            descriptor_schemas.extend(
                self.find_schemas(desc_path, filters=schemas, schema_type='descriptor', verbosity=verbosity))

        self.register_descriptors(descriptor_schemas, user_admin, force, verbosity=verbosity)

        if verbosity > 0:
            self.stdout.write("Running executor post-registration hook...")
        manager.get_executor().post_register_hook(verbosity=verbosity)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import sys
import unittest

from django.contrib.auth.models import AnonymousUser
from django.test import LiveServerTestCase

from guardian.shortcuts import assign_perm

from resolwe.flow.models import Collection, Process, Storage
from resolwe.test import ProcessTestCase, check_installed, tag_process, with_docker_executor, with_resolwe_host


class AccessAPIFromExecutorProcessTestCase(ProcessTestCase, LiveServerTestCase):

    def _create_collection(self):
        """Create a test collection that will be accessed via a test process.

        Assign :class:`~django.contrib.auth.models.AnonymousUser`
        ``view_collection`` permission to the created collection.

        :return: created test collection
        :rtype: Collection

        """
        collection = Collection.objects.create(
            name='collection-foo',
            contributor=self.contributor,
        )
        assign_perm("view_collection", AnonymousUser(), collection)
        return collection

    def create_process(self):
        """Create a test process that will query the API for collections.

        Get the list of all collections using ``curl`` to query the Live
        Resolwe host.

        :return: created test process
        :rtype: Process

        """
        process = Process.objects.create(
            name='Test accessing API from process',
            requirements={'expression-engine': 'jinja', 'resources': {'network': True}},
            contributor=self.contributor,
            type='data:test:api-access',
            input_schema=[],
            output_schema=[
                {
                    'name': 'collection-list',
                    'type': 'basic:json:',
                    'required': False,
                },
            ],
            run={
                'language': 'bash',
                # NOTE: Resolwe Runtime Utilities are not available in the
                # resolwe/test:base Docker image so we need to use plain echo
                # NOTE: curl must be called with --silent argument to not ouput
                # its progress since it can interfere with parsing of JSON on
                # process' standard output
                'program': r"""
echo "{\"collection-list\": $(curl --silent --show-error $RESOLWE_HOST_URL/collection)}"
"""
            }
        )
        return process

    def setUp(self):
        """Custom initilization of :class:`~ProcessTestCase`."""
        super(AccessAPIFromExecutorProcessTestCase, self).setUp()

        self.process = self.create_process()

    def check_results(self, data):
        """Check if information obtained for collection via API is correct."""
        if 'collection-list' not in data.output:
            self.fail(
                "The test process didn't obtain collection list from the live Resolwe host." +
                self._debug_info(data)
            )
        collection_list = Storage.objects.get(data=data).json
        self.assertEqual(len(collection_list), 1)
        self.assertEqual(collection_list[0]['slug'], self.collection.slug)

    @unittest.skipUnless(*check_installed('curl'))
    @with_resolwe_host
    @tag_process('test-accessing-api-from-process')
    def test_access_api_from_local_executor_process(self):  # pylint: disable=invalid-name
        """Test if a process running via local executor can access API."""
        data = self.run_process(self.process.slug)
        self.check_results(data)

    @unittest.skipUnless(*check_installed('curl'))
    @tag_process('test-accessing-api-from-process')
    def test_access_api_from_local_executor_process_without_decorator(self):  # pylint: disable=invalid-name
        """Test if a process running via local executor cannot access API."""
        data = self.run_process(self.process.slug)
        with self.assertRaises(AssertionError):
            self.check_results(data)

    @unittest.skipUnless(
        sys.platform.startswith('linux'),
        "Accessing live Resolwe host from a Docker container on non-Linux systems is not possible yet."
    )
    @with_resolwe_host
    @with_docker_executor
    @tag_process('test-accessing-api-from-process')
    def test_access_api_from_docker_executor_process(self):
        """Test if a process running via Docker executor can access API."""
        # pylint: disable=invalid-name
        # NOTE: pylint gets confused if the above string is inline with the method name
        data = self.run_process(self.process.slug)
        self.check_results(data)

    @unittest.skipUnless(
        sys.platform.startswith('linux'),
        "Accessing live Resolwe host from a Docker container on non-Linux systems is not possible yet."
    )
    @with_docker_executor
    @with_resolwe_host
    @tag_process('test-accessing-api-from-process')
    def test_access_api_from_docker_executor_process_reverse_decorators(self):
        """Test if a process running via Docker executor can access API."""
        # pylint: disable=invalid-name
        # NOTE: pylint gets confused if the above string is inline with the method name
        data = self.run_process(self.process.slug)
        self.check_results(data)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.8 on 2016-10-13 09:15
from __future__ import unicode_literals

from django.db import migrations

from resolwe.flow.utils import get_data_checksum


def calculate_checksum(apps, schema_editor):
    Data = apps.get_model("flow", "Data")
    for data in Data.objects.all():
        data.checksum = get_data_checksum(data.input, data.process.slug, data.process.version)
        data.save()


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0016_update_versionfield'),
    ]

    operations = [
        migrations.RunPython(calculate_checksum),
    ]



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from resolwe.flow.models import Process
from resolwe.test import TestCase


class ManagerTest(TestCase):

    def setUp(self):
        super(ManagerTest, self).setUp()

        self.data = {'name': 'Test Process',
                     'contributor': self.contributor,
                     'type': 'data:test',
                     'version': 1}

    def test_slug(self):
        p = Process.objects.create(**self.data)
        self.assertEqual(p.slug, 'test-process')

        self.data['version'] = 2
        p = Process.objects.create(**self.data)
        self.assertEqual(p.slug, 'test-process')

        p = Process.objects.create(**self.data)
        self.assertEqual(p.slug, 'test-process-2')



"""Resolwe process schema migration operations."""
import inspect

import jsonschema

from django.db.migrations.operations import base

from resolwe.flow.models.utils import validate_process_types, validation_schema
from resolwe.flow.utils import dict_dot

FIELD_SCHEMA = validation_schema('field')


class DataDefaultOperation(object):
    """Abstract data default generator."""

    def prepare(self, data, from_state):
        """Prepare data object migration.

        Called after all relevant process schemas have already
        been migrated.

        :param data: Queryset containing all data objects that need
            to be migrated
        :param from_state: Database model state
        """
        raise NotImplementedError

    def get_default_for(self, data, from_state):
        """Return default for given data object.

        :param data: Data object instance
        :param from_state: Database model state
        """
        raise NotImplementedError


class ConstantDataDefault(DataDefaultOperation):
    """Data default, which generates constant values."""

    def __init__(self, value):
        """Initialize constant data default.

        :param value: Constant default value to be used for all
            Data objects
        """
        self.value = value

    def prepare(self, data, from_state):
        """Prepare data object migration.

        Called after all relevant process schemas have already
        been migrated.

        :param data: Queryset containing all data objects that need
            to be migrated
        :param from_state: Database model state
        """
        return

    def get_default_for(self, data, from_state):
        """Return default for given data object.

        :param data: Data object instance
        :param from_state: Database model state
        """
        return self.value


class ResolweProcessOperation(base.Operation):
    """Abstract resolwe process migration operation."""

    reduces_to_sql = False
    atomic = True
    reversible = False

    def __init__(self, process, schema_type, **kwargs):
        """
        Construct a process migration operation.

        :param process: Process slug
        :param schema_type: Type of schema (must be either ``input`` or
            ``output``)
        """
        self.process = process
        self.schema_type = schema_type
        if self.schema_type not in ('input', 'output'):
            raise ValueError("Schema type must be either input or output")

        # Unfortunately, we can't easily get the identifier of the current
        # migration, which we need so we can update migration history. Here,
        # we assume that the Python implementation supports inspecting the
        # stack and we extract the parent migration's __module__ from there.
        migration_id = None
        current_frame = inspect.currentframe().f_back
        while current_frame:
            info = inspect.getframeinfo(current_frame)
            if info.function != 'Migration':
                current_frame = current_frame.f_back
                continue

            migration_id = current_frame.f_locals['__module__']
            break

        if not migration_id:
            raise ValueError("Unable to determine parent migration")

        self.migration_id = migration_id

    def deconstruct(self):
        """Deconstruct operation."""
        return (
            self.__class__.__name__,
            [],
            {
                'process': self.process,
                'schema_type': self.schema_type,
            }
        )

    def state_forwards(self, app_label, state):
        """Operation does not alter database state, only data."""
        pass

    def migrate_process_schema(self, process, schema, from_state):
        """Migrate process schema.

        :param process: Process instance
        :param schema: Process schema to migrate
        :param from_state: Database model state
        :return: True if the process was migrated, False otherwise
        """
        raise NotImplementedError

    def migrate_data(self, data, from_state):
        """Migrate data objects.

        :param data: Queryset containing all data objects that need
            to be migrated
        :param from_state: Database model state
        """
        raise NotImplementedError

    def describe_process_migration(self, process):
        """Return process migration metadata.

        These metadata will be recorded in the migration history and
        must be JSON-serializable. The default implementation returns
        an empty dictionary.

        :param process: Migrated Process object instance
        """
        return {}

    def describe_data_migration(self, data):
        """Return data migration metadata.

        These metadata will be recorded in the migration history and
        must be JSON-serializable. The default implementation returns
        an empty dictionary.

        :param data: Migrated Data object instance
        """
        return {}

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        """Perform forward migration."""
        Process = from_state.apps.get_model('flow', 'Process')  # pylint: disable=invalid-name
        Data = from_state.apps.get_model('flow', 'Data')  # pylint: disable=invalid-name
        try:
            # pylint: disable=invalid-name
            ProcessMigrationHistory = from_state.apps.get_model('flow', 'ProcessMigrationHistory')
            DataMigrationHistory = from_state.apps.get_model('flow', 'DataMigrationHistory')
        except LookupError:
            raise LookupError(
                "Unable to retrieve migration history models. Perhaps you need "
                "to add a migration dependency to a recent enough Resolwe flow "
                "app in your migration?"
            )

        # Migrate processes.
        processes = Process.objects.filter(slug=self.process)
        if not processes.exists():
            return

        migrated_processes = set()
        schema_field = '{}_schema'.format(self.schema_type)
        for process in processes:
            current_schema = getattr(process, schema_field)
            if not self.migrate_process_schema(process, current_schema, from_state):
                continue

            setattr(process, schema_field, current_schema)
            process.save()
            migrated_processes.add(process)

            # Update process migration log.
            ProcessMigrationHistory.objects.create(
                migration=self.migration_id,
                process=process,
                metadata=self.describe_process_migration(process),
            )

        if not migrated_processes:
            return

        # Migrate all data objects.
        data = Data.objects.filter(process__in=migrated_processes)
        self.migrate_data(data, from_state)

        # Update data migration log.
        for datum in data:
            DataMigrationHistory.objects.create(
                migration=self.migration_id,
                data=datum,
                metadata=self.describe_data_migration(datum)
            )

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        """Backward migration not possible."""
        raise NotImplementedError


class ResolweProcessAddField(ResolweProcessOperation):  # pylint: disable=abstract-method
    """Operation, which adds a new field to existing process schema."""

    def __init__(self, process, field, schema, default=None):
        """Construct a process field add operation.

        :param process: Process slug
        :param field: Dot-separated field path, starting with the top-level
            database field name (must be either ``input`` or ``output``)
        :param schema: Field schema definition
        :param default: An operation that generates default values, which
            must be an instance of :class:`DataDefaultOperation` or its
            subclass
        """
        self._raw_field = field
        self.field = field.split('.')
        if self.field[0] not in ('input', 'output'):
            raise ValueError("Field path must start with either input or output")
        if len(self.field) < 2:
            raise ValueError("Field path must contain at least two levels")
        if len(self.field) != 2:
            # TODO: Support nested fields.
            raise NotImplementedError("Field path must contain exactly two levels")
        schema_type = self.field.pop(0)

        self.schema = schema
        jsonschema.validate([schema], FIELD_SCHEMA)

        if schema['name'] != self.field[-1]:
            raise ValueError("Field name in schema differs from field path")

        if not default and schema.get('required', True):
            raise ValueError("A required field must have a default")
        if default and not isinstance(default, DataDefaultOperation):
            raise TypeError("Default must be an instance of DataDefaultOperation")
        self.default = default

        super().__init__(
            process=process,
            schema_type=schema_type,
            field=field,
            schema=schema,
            default=default
        )

    def deconstruct(self):
        """Deconstruct operation."""
        return (
            self.__class__.__name__,
            [],
            {
                'process': self.process,
                'field': self._raw_field,
                'schema': self.schema,
                'default': self.default,
            }
        )

    def migrate_process_schema(self, process, schema, from_state):
        """Migrate process schema.

        :param process: Process instance
        :param schema: Process schema to migrate
        :param from_state: Database model state
        :return: True if the process was migrated, False otherwise
        """
        container = dict_dot(schema, '.'.join(self.field[:-1]), default=list)

        # Ignore processes, which already contain the target field with the
        # target schema.
        for field in container:
            if field['name'] == self.field[-1]:
                if field == self.schema:
                    return False
                else:
                    raise ValueError(
                        "Failed to migrate schema for process '{process}' as the field '{field}' "
                        "already exists and has an incompatible schema".format(
                            process=process.slug,
                            field=self.field[-1]
                        )
                    )

        # Add field to container.
        container.append(self.schema)
        return True

    def migrate_data(self, data, from_state):
        """Migrate data objects.

        :param data: Queryset containing all data objects that need
            to be migrated
        :param from_state: Database model state
        """
        if not self.default:
            return

        self.default.prepare(data, from_state)
        for instance in data:
            value = self.default.get_default_for(instance, from_state)
            if not value and not self.schema.get('required', True):
                continue

            # Set default value.
            container = getattr(instance, self.schema_type, {})
            dict_dot(container, '.'.join(self.field), value)
            setattr(instance, self.schema_type, container)
            instance.save()

    def describe(self):
        """Migration description."""
        return "Add Resolwe process field"

    def describe_process_migration(self, process):
        """Return process migration metadata.

        These metadata will be recorded in the migration history and
        must be JSON-serializable. The default implementation returns
        an empty dictionary.

        :param process: Migrated Process object instance
        """
        return {
            'operation': 'add_field',
            'field': self._raw_field,
        }

    def describe_data_migration(self, data):
        """Return data migration metadata.

        These metadata will be recorded in the migration history and
        must be JSON-serializable. The default implementation returns
        an empty dictionary.

        :param data: Migrated Data object instance
        """
        return {
            'operation': 'add_field',
            'field': self._raw_field,
        }


class ResolweProcessRenameField(ResolweProcessOperation):  # pylint: disable=abstract-method
    """Operation, which renames an existing field in existing process schema."""

    def __init__(self, process, field, new_field, skip_no_field=False):
        """Construct a process field rename operation.

        :param process: Process slug
        :param field: Dot-separated field path, starting with the top-level
            database field name (must be either ``input`` or ``output``)
        :param new_field: New field name
        """
        self._raw_field = field
        self.field = field.split('.')
        if self.field[0] not in ('input', 'output'):
            raise ValueError("Field path must start with either input or output")
        if len(self.field) < 2:
            raise ValueError("Field path must contain at least two levels")
        if len(self.field) != 2:
            # TODO: Support nested fields.
            raise NotImplementedError("Field path must contain exactly two levels")

        schema_type = self.field.pop(0)

        self.new_field = new_field
        self.skip_no_field = skip_no_field

        super().__init__(
            process=process,
            schema_type=schema_type,
        )

    def deconstruct(self):
        """Deconstruct operation."""
        return (
            self.__class__.__name__,
            [],
            {
                'process': self.process,
                'field': self._raw_field,
                'new_field': self.new_field,
            }
        )

    def migrate_process_schema(self, process, schema, from_state):
        """Migrate process schema.

        :param process: Process instance
        :param schema: Process schema to migrate
        :param from_state: Database model state
        :return: True if the process was migrated, False otherwise
        """
        container = dict_dot(schema, '.'.join(self.field[:-1]), default=list)

        # Ignore processes, which already contain the target field.
        migrate = False
        for field in container:
            if field['name'] == self.field[-1]:
                field['name'] = self.new_field
                migrate = True
                break
            elif field['name'] == self.new_field:
                # Already has target field.
                migrate = False
                break
        else:
            if not self.skip_no_field:
                raise ValueError(
                    "Unable to rename: there is no field with name '{field}' or '{new_field}'.".format(
                        field=self.field[-1],
                        new_field=self.new_field,
                    )
                )

        return migrate

    def migrate_data(self, data, from_state):
        """Migrate data objects.

        :param data: Queryset containing all data objects that need
            to be migrated
        :param from_state: Database model state
        """
        for instance in data:
            if instance.status == 'ER':
                continue

            container = getattr(instance, self.schema_type, {})
            schema = container.pop(self.field[-1])
            container[self.new_field] = schema

            setattr(instance, self.schema_type, container)
            instance.save()

    def describe(self):
        """Migration description."""
        return "Rename Resolwe process field"

    def describe_process_migration(self, process):
        """Return process migration metadata.

        These metadata will be recorded in the migration history and
        must be JSON-serializable. The default implementation returns
        an empty dictionary.

        :param process: Migrated Process object instance
        """
        return {
            'operation': 'rename_field',
            'field': self._raw_field,
            'new_field': self.new_field,
        }

    def describe_data_migration(self, data):
        """Return data migration metadata.

        These metadata will be recorded in the migration history and
        must be JSON-serializable. The default implementation returns
        an empty dictionary.

        :param data: Migrated Data object instance
        """
        return {
            'operation': 'rename_field',
            'field': self._raw_field,
            'new_field': self.new_field,
        }


class ResolweDataCleanup(base.Operation):
    """Resolwe temporary and errored Data object removal operation."""

    reduces_to_sql = False
    atomic = True
    reversible = False

    def state_forwards(self, app_label, state):
        """Operation does not alter database state, only data."""
        pass

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        """Perform forward migration."""
        from resolwe.flow.models.data import DataQuerySet
        # pylint: disable=protected-access

        Data = from_state.apps.get_model('flow', 'Data')  # pylint: disable=invalid-name
        DataQuerySet._delete_chunked(Data.objects.filter(process__persistence='TMP'))
        DataQuerySet._delete_chunked(Data.objects.filter(status='ER'))

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        """Backward migration not possible."""
        raise NotImplementedError


class ResolweProcessDataRemove(base.Operation):
    """Resolwe remove Process and Data operation."""

    reduces_to_sql = False
    atomic = True
    reversible = False

    def __init__(self, process):
        """
        Remove a Process and all of it's Data.

        :param process: Process slug
        """
        self.process = process

    def state_forwards(self, app_label, state):
        """Operation does not alter database state, only data."""
        pass

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        """Perform forward migration."""
        from resolwe.flow.models.data import DataQuerySet
        # pylint: disable=protected-access

        Data = from_state.apps.get_model('flow', 'Data')  # pylint: disable=invalid-name
        DataQuerySet._delete_chunked(Data.objects.filter(process__slug=self.process))

        Process = from_state.apps.get_model('flow', 'Process')  # pylint: disable=invalid-name
        Process.objects.filter(slug=self.process).delete()

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        """Backward migration not possible."""
        raise NotImplementedError


class ResolweProcessChangeType(ResolweProcessOperation):  # pylint: disable=abstract-method
    """Operation, changes a process' type."""

    def __init__(self, process, new_type):
        """Construct a process type change operation.

        :param process: Process slug
        :param new_type: New process data type
        """
        self.new_type = new_type
        # TODO: Validate type.

        super().__init__(
            process=process,
            schema_type='output',
        )

    def deconstruct(self):
        """Deconstruct operation."""
        return (
            self.__class__.__name__,
            [],
            {
                'process': self.process,
                'new_type': self.new_type,
            }
        )

    def migrate_process_schema(self, process, schema, from_state):
        """Migrate process schema.

        :param process: Process instance
        :param schema: Process schema to migrate
        :param from_state: Database model state
        :return: True if the process was migrated, False otherwise
        """
        if process.type == self.new_type:
            return False

        process.type = self.new_type
        return True

    def migrate_data(self, data, from_state):
        """Migrate data objects.

        :param data: Queryset containing all data objects that need
            to be migrated
        :param from_state: Database model state
        """
        pass

    def describe(self):
        """Migration description."""
        return "Change Resolwe process type"

    def describe_process_migration(self, process):
        """Return process migration metadata.

        These metadata will be recorded in the migration history and
        must be JSON-serializable. The default implementation returns
        an empty dictionary.

        :param process: Migrated Process object instance
        """
        return {
            'operation': 'change_type',
            'new_type': self.new_type,
        }

    def describe_data_migration(self, data):
        """Return data migration metadata.

        These metadata will be recorded in the migration history and
        must be JSON-serializable. The default implementation returns
        an empty dictionary.

        :param data: Migrated Data object instance
        """
        return {
            'operation': 'change_type',
            'new_type': self.new_type,
        }


class ResolweValidateProcessSchema(base.Operation):
    """Process schema validation operation."""

    reduces_to_sql = False
    atomic = True
    reversible = False

    def state_forwards(self, app_label, state):
        """Operation does not alter database state, only data."""
        pass

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        """Perform forward migration."""
        Process = from_state.apps.get_model('flow', 'Process')  # pylint: disable=invalid-name

        # Validate process types to ensure consistency.
        errors = validate_process_types(Process.objects.all())
        if errors:
            raise ValueError("Process type consistency check failed after migration: {}".format(errors))

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        """Backward migration not possible."""
        raise NotImplementedError



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from mock import patch

from django.core.exceptions import ValidationError

from rest_framework.response import Response

from resolwe.flow.models import Data, Process
from resolwe.flow.utils import get_data_checksum
from resolwe.flow.utils.exceptions import resolwe_exception_handler
from resolwe.test import TestCase


class ExceptionsTestCase(TestCase):

    @patch('resolwe.flow.utils.exceptions.exception_handler')
    def test_exception_handler(self, exception_handler_mock):
        exception_handler_mock.return_value = None
        resp = resolwe_exception_handler(Exception, {})
        self.assertEqual(resp, None)

        exception_handler_mock.return_value = None
        resp = resolwe_exception_handler(ValidationError('Error description'), {})
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.data['error'], 'Error description')

        original_resp = Response({})
        exception_handler_mock.return_value = original_resp
        resp = resolwe_exception_handler(ValidationError('Error description'), {})
        self.assertEqual(id(original_resp), id(resp))
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.data['error'], 'Error description')

    def test_checksum_consistency(self):
        process = Process(version='1.0.0', slug='my-process')
        data = Data()

        data.input = {'tss': 0, 'genome': 'HG19'}
        checksum = get_data_checksum(data.input, process.slug, process.version)
        self.assertEqual(checksum, 'ca322c2bb48b58eea3946e624fe6cfdc53c2cc12478465b6f0ca2d722e280c4c')

        data.input = {'genome': 'HG19', 'tss': 0}
        checksum = get_data_checksum(data.input, process.slug, process.version)
        self.assertEqual(checksum, 'ca322c2bb48b58eea3946e624fe6cfdc53c2cc12478465b6f0ca2d722e280c4c')



# pylint: disable=missing-docstring
import os

from resolwe.test import ProcessTestCase, tag_process, with_docker_executor


class ArchiverProcessTestCase(ProcessTestCase):

    def setUp(self):
        super(ArchiverProcessTestCase, self).setUp()
        self.files_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'files'))

    @with_docker_executor
    @tag_process('archiver')
    def test_archiver(self):
        with self.preparation_stage():
            binary = self.run_process('upload-file', {'src': 'file binary'})
            image = self.run_process('upload-image-file', {'src': 'file image.png'})
            tab = self.run_process('upload-tab-file', {'src': 'file tab.txt'})

        archive = self.run_process('archiver', {
            'data': [binary.id, image.id, tab.id],
            'fields': ['file']})

        self.assertFileExists(archive, 'archive')
